"""Tests for iconfucius.cli.trade — buy/sell tokens on Odin.Fun."""

from unittest.mock import MagicMock, patch

import pytest

M = "iconfucius.cli.trade"


def _make_mock_auth(bot_principal="bot-principal-abc"):
    """Create a mock auth for testing."""
    delegate = MagicMock()
    delegate.der_pubkey = b"\x30" * 44
    return {
        "delegate_identity": delegate,
        "bot_principal_text": bot_principal,
        "jwt_token": "jwt",
    }


class TestRunTradeSuccess:
    @patch(f"{M}.get_btc_to_usd_rate", return_value=100_000.0)
    @patch(f"{M}.unwrap_canister_result", side_effect=lambda x: x)
    @patch(f"{M}.patch_delegate_sender")
    @patch(f"{M}.load_session")
    @patch(f"{M}._fetch_token_info", return_value={"ticker": "TEST", "price": 1000})
    @patch(f"{M}.Canister")
    @patch(f"{M}.Agent")
    @patch(f"{M}.Client")
    def test_buy(self, MockClient, MockAgent, MockCanister, mock_token_info,
                  mock_load, mock_patch_del, mock_unwrap, mock_rate,
                  odin_project):
        """Verify buy."""
        mock_load.return_value = _make_mock_auth()
        mock_odin = MagicMock()
        mock_odin.getBalance.side_effect = [5_000_000, 100]  # BTC msat, token
        mock_odin.token_trade.return_value = {"ok": None}
        MockCanister.side_effect = [mock_odin, mock_odin]

        from iconfucius.cli.trade import run_trade
        result = run_trade(bot_name="bot-1", action="buy", token_id="29m8",
                           amount="1000", verbose=False)

        assert result["status"] == "ok"
        assert result["action"] == "buy"
        assert result["bot_name"] == "bot-1"
        assert result["token_id"] == "29m8"
        assert result["amount"] == 1000
        assert "note" not in result
        mock_odin.token_trade.assert_called_once()

    @patch(f"{M}.get_btc_to_usd_rate", return_value=100_000.0)
    @patch(f"{M}.unwrap_canister_result", side_effect=lambda x: x)
    @patch(f"{M}.patch_delegate_sender")
    @patch(f"{M}.load_session")
    @patch(f"{M}._fetch_token_info", return_value={"ticker": "TEST", "price": 1000})
    @patch(f"{M}.Canister")
    @patch(f"{M}.Agent")
    @patch(f"{M}.Client")
    def test_buy_capped_to_balance(self, MockClient, MockAgent, MockCanister,
                                    mock_token_info, mock_load, mock_patch_del,
                                    mock_unwrap, mock_rate,
                                    odin_project):
        """Buy amount exceeding Odin.Fun balance is auto-capped."""
        mock_load.return_value = _make_mock_auth()
        mock_odin = MagicMock()
        mock_odin.getBalance.side_effect = [3_000_000, 0]  # 3000 sats on Odin
        mock_odin.token_trade.return_value = {"ok": None}
        MockCanister.side_effect = [mock_odin, mock_odin]

        from iconfucius.cli.trade import run_trade
        # Request 5000 sats -> should be capped to 3000
        result = run_trade(bot_name="bot-1", action="buy", token_id="29m8",
                           amount="5000", verbose=False)

        assert result["status"] == "ok"
        assert result["amount"] == 3000  # capped to balance
        assert "auto-capped" in result["note"]
        call_args = mock_odin.token_trade.call_args[0][0]
        assert call_args["amount"] == {"btc": 3_000_000}  # 3000 sats in msat

    @patch(f"{M}.get_btc_to_usd_rate", return_value=100_000.0)
    @patch(f"{M}.unwrap_canister_result", side_effect=lambda x: x)
    @patch(f"{M}.patch_delegate_sender")
    @patch(f"{M}.load_session")
    @patch(f"{M}._fetch_token_info", return_value={"ticker": "TEST", "price": 1000})
    @patch(f"{M}.Canister")
    @patch(f"{M}.Agent")
    @patch(f"{M}.Client")
    def test_buy_balance_below_minimum(self, MockClient, MockAgent, MockCanister,
                                        mock_token_info, mock_load, mock_patch_del,
                                        mock_unwrap, mock_rate,
                                        odin_project):
        """Buy fails when Odin.Fun balance is below MIN_TRADE_SATS."""
        mock_load.return_value = _make_mock_auth()
        mock_odin = MagicMock()
        mock_odin.getBalance.side_effect = [100_000, 0]  # 100 sats < 500 min
        MockCanister.side_effect = [mock_odin, mock_odin]

        from iconfucius.cli.trade import run_trade
        result = run_trade(bot_name="bot-1", action="buy", token_id="29m8",
                           amount="5000", verbose=False)

        assert result["status"] == "error"
        assert "too low" in result["error"]
        mock_odin.token_trade.assert_not_called()

    @patch(f"{M}.get_btc_to_usd_rate", return_value=100_000.0)
    @patch(f"{M}.unwrap_canister_result", side_effect=lambda x: x)
    @patch(f"{M}.patch_delegate_sender")
    @patch(f"{M}.load_session")
    @patch(f"{M}._fetch_token_info", return_value={"ticker": "TEST", "price": 500_000_000_000_000})
    @patch(f"{M}.Canister")
    @patch(f"{M}.Agent")
    @patch(f"{M}.Client")
    def test_sell(self, MockClient, MockAgent, MockCanister, mock_token_info,
                   mock_load, mock_patch_del, mock_unwrap, mock_rate,
                   odin_project):
        """Verify sell."""
        mock_load.return_value = _make_mock_auth()
        mock_odin = MagicMock()
        mock_odin.getBalance.side_effect = [5_000_000, 500]
        mock_odin.token_trade.return_value = {"ok": None}
        MockCanister.side_effect = [mock_odin, mock_odin]

        from iconfucius.cli.trade import run_trade
        result = run_trade(bot_name="bot-1", action="sell", token_id="29m8",
                           amount="100", verbose=False)

        assert result["status"] == "ok"
        assert result["action"] == "sell"
        assert result["amount"] == 100


class TestRunTradeSellAll:
    @patch(f"{M}.get_btc_to_usd_rate", return_value=100_000.0)
    @patch(f"{M}.unwrap_canister_result", side_effect=lambda x: x)
    @patch(f"{M}.patch_delegate_sender")
    @patch(f"{M}.load_session")
    @patch(f"{M}._fetch_token_info", return_value={"ticker": "TEST", "price": 500_000_000_000_000})
    @patch(f"{M}.Canister")
    @patch(f"{M}.Agent")
    @patch(f"{M}.Client")
    def test_sell_all(self, MockClient, MockAgent, MockCanister, mock_token_info,
                      mock_load, mock_patch_del, mock_unwrap, mock_rate,
                      odin_project):
        """Verify sell all."""
        mock_load.return_value = _make_mock_auth()
        mock_odin = MagicMock()
        mock_odin.getBalance.side_effect = [5_000_000, 99_999]
        mock_odin.token_trade.return_value = {"ok": None}
        MockCanister.side_effect = [mock_odin, mock_odin]

        from iconfucius.cli.trade import run_trade
        result = run_trade(bot_name="bot-1", action="sell", token_id="29m8",
                           amount="all", verbose=False)

        assert result["status"] == "ok"
        assert result["amount"] == 99_999
        # Verify trade used the full token balance
        call_args = mock_odin.token_trade.call_args[0][0]
        assert call_args["amount"] == {"token": 99_999}

    @patch(f"{M}.get_btc_to_usd_rate", return_value=100_000.0)
    @patch(f"{M}.unwrap_canister_result", side_effect=lambda x: x)
    @patch(f"{M}.patch_delegate_sender")
    @patch(f"{M}.load_session")
    @patch(f"{M}._fetch_token_info", return_value={"ticker": "TEST", "price": 1000})
    @patch(f"{M}.Canister")
    @patch(f"{M}.Agent")
    @patch(f"{M}.Client")
    def test_sell_all_zero_balance(self, MockClient, MockAgent, MockCanister,
                                    mock_token_info, mock_load, mock_patch_del,
                                    mock_unwrap, mock_rate,
                                    odin_project):
        """Verify sell all zero balance."""
        mock_load.return_value = _make_mock_auth()
        mock_odin = MagicMock()
        mock_odin.getBalance.side_effect = [5_000_000, 0]
        MockCanister.side_effect = [mock_odin, mock_odin]

        from iconfucius.cli.trade import run_trade
        result = run_trade(bot_name="bot-1", action="sell", token_id="29m8",
                           amount="all", verbose=False)

        assert result["status"] == "skipped"
        assert "to sell" in result["reason"]
        mock_odin.token_trade.assert_not_called()


class TestRunTradeErrors:
    def test_no_wallet(self, odin_project_no_wallet):
        """Verify no wallet."""
        from iconfucius.cli.trade import run_trade
        result = run_trade(bot_name="bot-1", action="buy", token_id="29m8", amount="1000")
        assert result["status"] == "error"
        assert "wallet" in result["error"].lower()

    def test_invalid_action(self, odin_project):  # noqa: ARG002
        """Verify invalid action."""
        from iconfucius.cli.trade import run_trade
        result = run_trade(bot_name="bot-1", action="hold", token_id="29m8", amount="1000")
        assert result["status"] == "error"
        assert "must be 'buy' or 'sell'" in result["error"]

    def test_buy_all_rejected(self, odin_project):  # noqa: ARG002
        """Verify buy all rejected."""
        from iconfucius.cli.trade import run_trade
        result = run_trade(bot_name="bot-1", action="buy", token_id="29m8", amount="all")
        assert result["status"] == "error"
        assert "only supported for sell" in result["error"]

    @patch(f"{M}.get_btc_to_usd_rate", return_value=100_000.0)
    @patch(f"{M}.unwrap_canister_result", side_effect=lambda x: x)
    @patch(f"{M}.patch_delegate_sender")
    @patch(f"{M}.load_session")
    @patch(f"{M}._fetch_token_info", return_value={"ticker": "TEST", "price": 1000})
    @patch(f"{M}.Canister")
    @patch(f"{M}.Agent")
    @patch(f"{M}.Client")
    def test_trade_failure(self, MockClient, MockAgent, MockCanister,
                            mock_token_info, mock_load, mock_patch_del,
                            mock_unwrap, mock_rate,
                            odin_project):
        """Verify trade failure."""
        mock_load.return_value = _make_mock_auth()
        mock_odin = MagicMock()
        mock_odin.getBalance.side_effect = [5_000_000, 100]
        mock_odin.token_trade.return_value = {"err": "insufficient BTC"}
        MockCanister.side_effect = [mock_odin, mock_odin]

        from iconfucius.cli.trade import run_trade
        result = run_trade(bot_name="bot-1", action="buy", token_id="29m8",
                           amount="1000", verbose=False)

        assert result["status"] == "error"
        assert "insufficient BTC" in result["error"]

    @patch(f"{M}.get_btc_to_usd_rate", return_value=100_000.0)
    @patch(f"{M}.unwrap_canister_result", side_effect=lambda x: x)
    @patch(f"{M}.patch_delegate_sender")
    @patch(f"{M}.load_session")
    @patch(f"{M}._fetch_token_info", return_value=None)
    @patch(f"{M}.Canister")
    @patch(f"{M}.Agent")
    @patch(f"{M}.Client")
    def test_token_info_unavailable(self, MockClient, MockAgent, MockCanister,
                                     mock_token_info, mock_load, mock_patch_del,
                                     mock_unwrap, mock_rate,
                                     odin_project):
        """Trade should work even if token info API is unavailable."""
        mock_load.return_value = _make_mock_auth()
        mock_odin = MagicMock()
        mock_odin.getBalance.side_effect = [5_000_000, 100]
        mock_odin.token_trade.return_value = {"ok": None}
        MockCanister.side_effect = [mock_odin, mock_odin]

        from iconfucius.cli.trade import run_trade

        # Should not raise
        result = run_trade(bot_name="bot-1", action="buy", token_id="29m8",
                           amount="1000", verbose=False)

        assert result["status"] == "ok"



"""Tests for LLM hallucination scenarios.

Simulates cases where the AI generates tool calls with wrong, dangerous,
or nonsensical parameters. Each scenario is defined in a list so real
hallucinations from production conversations can be easily added.

The tests verify that:
1. The confirmation prompt shows accurate info (so the user can catch it)
2. Handlers reject clearly invalid inputs
3. Pre-conversion doesn't silently corrupt amounts
"""

from unittest.mock import MagicMock, patch

import pytest

from iconfucius.cli.chat import _describe_tool_call, _run_tool_loop
from iconfucius.skills.executor import execute_tool


# ---------------------------------------------------------------------------
# Helper
# ---------------------------------------------------------------------------

def _make_tool_response(name, tool_input, *, tool_id="id_1"):
    """Build a mock LLM response containing one tool_use block."""
    tool_block = MagicMock()
    tool_block.type = "tool_use"
    tool_block.id = tool_id
    tool_block.name = name
    tool_block.input = dict(tool_input)

    text_block = MagicMock()
    text_block.type = "text"
    text_block.text = "Done."

    resp_tool = MagicMock()
    resp_tool.content = [tool_block]
    resp_text = MagicMock()
    resp_text.content = [text_block]
    return resp_tool, resp_text


# ===================================================================
# Hallucination scenarios — extend this list with real examples!
# ===================================================================

# Each entry:
#   id           - short unique id for pytest parametrize
#   description  - what the hallucination is
#   tool_name    - tool the AI called
#   tool_input   - hallucinated parameters
#   user_said    - what the user actually asked (for documentation)
#   expect_error - if True, handler must return status=error
#   expect_in_confirm - substring that MUST appear in confirmation prompt
#                       (so user can catch the mistake)
#   expect_not_in_confirm - substring that must NOT appear (optional)

HALLUCINATIONS = [
    # ---- Amount confusion ----
    {
        "id": "fund_10_sats_instead_of_10_usd",
        "description": "User says '$10', AI sends amount=10 (sats, not dollars)",
        "tool_name": "fund",
        "tool_input": {"bot_name": "bot-1", "amount": 10},
        "user_said": "fund bot-1 with $10",
        "expect_error": False,
        "expect_in_confirm": "10 sats",
    },
    {
        "id": "send_btc_not_sats",
        "description": "AI sends 0.001 (BTC) instead of 100,000 sats",
        "tool_name": "wallet_send",
        "tool_input": {"amount": "0.001", "address": "bc1qtest"},
        "user_said": "send 0.001 BTC to bc1qtest",
        "expect_error": False,
        "expect_in_confirm": "0.001",
    },
    {
        "id": "fund_negative_amount",
        "description": "AI hallucinates a negative amount",
        "tool_name": "fund",
        "tool_input": {"bot_name": "bot-1", "amount": -5000},
        "user_said": "fund bot-1 with 5000 sats",
        "expect_error": False,
        "expect_in_confirm": "-5,000",
    },
    {
        "id": "trade_buy_zero_sats",
        "description": "AI sends zero-amount buy order",
        "tool_name": "trade_buy",
        "tool_input": {"token_id": "29m8", "amount": 0, "bot_name": "bot-1"},
        "user_said": "buy $5 of IConfucius",
        "expect_error": True,
    },
    {
        "id": "withdraw_absurdly_large",
        "description": "AI hallucinates 1 BTC withdraw (100M sats)",
        "tool_name": "withdraw",
        "tool_input": {"bot_name": "bot-1", "amount": "100000000"},
        "user_said": "withdraw 1000 sats from bot-1",
        "expect_error": False,
        "expect_in_confirm": "100,000,000",
    },

    # ---- Wrong target ----
    {
        "id": "send_to_own_address",
        "description": "AI sends funds to the user's own wallet principal",
        "tool_name": "wallet_send",
        "tool_input": {"amount": "5000",
                       "address": "y7paa-3ewsi-2iqfz-xlcd7-jpcvq-ibor6"},
        "user_said": "send 5000 sats to someone else",
        "expect_error": False,
        "expect_in_confirm": "y7paa",
    },
    {
        "id": "fund_wrong_bot",
        "description": "AI picks bot-2 when user said bot-1",
        "tool_name": "fund",
        "tool_input": {"bot_name": "bot-2", "amount": 5000},
        "user_said": "fund bot-1 with 5000 sats",
        "expect_error": False,
        "expect_in_confirm": "bot-2",
    },
    {
        "id": "trade_nonexistent_bot",
        "description": "AI invents a bot name that doesn't exist",
        "tool_name": "trade_buy",
        "tool_input": {"token_id": "29m8", "amount": 5000,
                       "bot_name": "bot-99"},
        "user_said": "buy 5000 sats of IConfucius",
        "expect_error": False,
        "expect_in_confirm": "bot-99",
    },

    # ---- Wrong token ----
    {
        "id": "buy_wrong_token_id",
        "description": "AI uses wrong token ID",
        "tool_name": "trade_buy",
        "tool_input": {"token_id": "xxxx", "amount": 5000,
                       "bot_name": "bot-1"},
        "user_said": "buy 5000 sats of IConfucius (29m8)",
        "expect_error": False,
        "expect_in_confirm": "xxxx",
    },
    {
        "id": "sell_wrong_token_id",
        "description": "AI sells wrong token",
        "tool_name": "trade_sell",
        "tool_input": {"token_id": "wrong", "amount": "all",
                       "bot_name": "bot-1"},
        "user_said": "sell all my IConfucius tokens",
        "expect_error": False,
        "expect_in_confirm": "wrong",
    },

    # ---- Missing required fields ----
    {
        "id": "fund_no_bot",
        "description": "AI forgets to specify which bot to fund",
        "tool_name": "fund",
        "tool_input": {"amount": 5000},
        "user_said": "fund bot-1 with 5000 sats",
        "expect_error": True,
    },
    {
        "id": "send_no_address",
        "description": "AI forgets the destination address",
        "tool_name": "wallet_send",
        "tool_input": {"amount": "5000"},
        "user_said": "send 5000 sats to bc1qtest",
        "expect_error": True,
    },
    {
        "id": "trade_buy_no_token",
        "description": "AI forgets token_id for buy",
        "tool_name": "trade_buy",
        "tool_input": {"amount": 5000, "bot_name": "bot-1"},
        "user_said": "buy 5000 sats of IConfucius",
        "expect_error": True,
    },
    {
        "id": "trade_sell_no_token",
        "description": "AI forgets token_id for sell",
        "tool_name": "trade_sell",
        "tool_input": {"amount": "all", "bot_name": "bot-1"},
        "user_said": "sell all IConfucius",
        "expect_error": True,
    },

    # ---- Confused operations ----
    {
        "id": "withdraw_sends_all_bots",
        "description": "User wants to withdraw from bot-1, AI sends all_bots",
        "tool_name": "withdraw",
        "tool_input": {"all_bots": True, "amount": "all"},
        "user_said": "withdraw all from bot-1",
        "expect_error": False,
        "expect_in_confirm": "all bots",
    },
    {
        "id": "fund_all_when_user_said_one",
        "description": "User funds one bot, AI funds all",
        "tool_name": "fund",
        "tool_input": {"all_bots": True, "amount": 5000},
        "user_said": "fund bot-1 with 5000 sats",
        "expect_error": False,
        "expect_in_confirm": "all bots",
        "expect_not_in_confirm": None,
    },

    # ---- USD/sats confusion ----
    {
        "id": "amount_usd_when_user_said_sats",
        "description": "User says 5000 sats, AI uses amount_usd=5000",
        "tool_name": "fund",
        "tool_input": {"bot_name": "bot-1", "amount_usd": 5000.0},
        "user_said": "fund bot-1 with 5000 sats",
        "expect_error": False,
        # After pre-conversion: $5000 at any rate → huge sats number
        # The confirm prompt must show this so user catches it
    },

    # ---- Sell all when user said sell some ----
    {
        "id": "sell_all_when_partial",
        "description": "User says 'sell 100 tokens', AI sends amount='all'",
        "tool_name": "trade_sell",
        "tool_input": {"token_id": "29m8", "amount": "all",
                       "bot_name": "bot-1"},
        "user_said": "sell 100 IConfucius tokens",
        "expect_error": False,
        "expect_in_confirm": "all",
    },

    # ---- Empty / whitespace address ----
    {
        "id": "send_whitespace_address",
        "description": "AI sends to whitespace-only address",
        "tool_name": "wallet_send",
        "tool_input": {"amount": "5000", "address": "   "},
        "user_said": "send 5000 sats to bc1qtest",
        "expect_error": True,
    },

    # --- Add real hallucinations from production below this line ---
    # {
    #     "id": "real_hallucination_001",
    #     "description": "Describe what happened",
    #     "tool_name": "...",
    #     "tool_input": {...},
    #     "user_said": "what the user actually typed",
    #     "expect_error": True/False,
    #     "expect_in_confirm": "...",
    # },
]


# ===================================================================
# Test: confirmation prompt shows accurate info for each hallucination
# ===================================================================

_CONFIRM_CASES = [h for h in HALLUCINATIONS if not h.get("expect_error")]


@pytest.mark.parametrize(
    "scenario",
    _CONFIRM_CASES,
    ids=[h["id"] for h in _CONFIRM_CASES],
)
def test_confirm_prompt_shows_hallucinated_values(scenario):
    """The confirm prompt must show the ACTUAL values the AI sent.

    This is the user's last line of defense — if the AI hallucinates
    a wrong amount/target/token, the confirmation prompt must make it
    visible so the user can decline.
    """
    desc = _describe_tool_call(scenario["tool_name"], scenario["tool_input"])
    expected = scenario.get("expect_in_confirm")
    if expected:
        assert expected in desc, (
            f"Confirmation prompt for '{scenario['id']}' should contain "
            f"'{expected}' but got: {desc}"
        )
    not_expected = scenario.get("expect_not_in_confirm")
    if not_expected:
        assert not_expected not in desc, (
            f"Confirmation prompt for '{scenario['id']}' should NOT contain "
            f"'{not_expected}' but got: {desc}"
        )


# ===================================================================
# Test: handlers reject clearly invalid hallucinated inputs
# ===================================================================

_ERROR_CASES = [h for h in HALLUCINATIONS if h.get("expect_error")]


@pytest.mark.parametrize(
    "scenario",
    _ERROR_CASES,
    ids=[h["id"] for h in _ERROR_CASES],
)
def test_handler_rejects_invalid_hallucination(scenario):
    """Handlers must return error for hallucinated invalid inputs."""
    with patch("iconfucius.config.require_wallet", return_value=True):
        result = execute_tool(scenario["tool_name"], scenario["tool_input"])
    assert result["status"] == "error", (
        f"Handler for '{scenario['id']}' should return error but got: {result}"
    )


# ===================================================================
# Test: pre-conversion makes hallucinated USD amounts visible
# ===================================================================

class TestPreConversionExposesHallucinations:
    """When AI hallucinates amount_usd, pre-conversion converts it to sats.

    This makes the hallucination visible in the confirmation prompt
    because fmt_sats shows "X sats ($Y.YY)" — the user sees the
    actual dollar amount and can catch mistakes.
    """

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100_000.0)
    @patch("iconfucius.cli.chat.execute_tool", return_value={"status": "ok"})
    @patch("iconfucius.cli.chat.get_tool_metadata",
           return_value={"requires_confirmation": True})
    def test_usd_5000_converted_to_huge_sats(self, mock_meta, mock_exec,
                                               mock_rate):
        """AI sends amount_usd=5000 (user said 5000 sats).

        Pre-conversion: $5000 / $100k * 1e8 = 5,000,000 sats.
        The confirm prompt shows "5,000,000 sats ($5,000.00)" — a
        clearly wrong amount that the user can decline.
        """
        backend = MagicMock()
        resp_tool, resp_text = _make_tool_response(
            "fund", {"bot_name": "bot-1", "amount_usd": 5000.0})
        backend.chat_with_tools.side_effect = [resp_tool, resp_text]

        with patch("builtins.input", return_value="y"):
            _run_tool_loop(backend, [], "system", [], "TestBot")

        call_args = mock_exec.call_args[0]
        # $5000 at $100k/BTC = 5,000,000 sats — clearly wrong
        assert call_args[1]["amount"] == 5_000_000
        assert "amount_usd" not in call_args[1]

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100_000.0)
    @patch("iconfucius.cli.chat.execute_tool", return_value={"status": "ok"})
    @patch("iconfucius.cli.chat.get_tool_metadata",
           return_value={"requires_confirmation": True})
    def test_usd_0_01_converted_shows_tiny_amount(self, mock_meta, mock_exec,
                                                    mock_rate):
        """AI sends amount_usd=0.01 when user said $10.

        Pre-conversion: $0.01 / $100k * 1e8 = 10 sats.
        Confirm shows "10 sats ($0.01)" — user sees wrong dollar amount.
        """
        backend = MagicMock()
        resp_tool, resp_text = _make_tool_response(
            "fund", {"bot_name": "bot-1", "amount_usd": 0.01})
        backend.chat_with_tools.side_effect = [resp_tool, resp_text]

        with patch("builtins.input", return_value="y"):
            _run_tool_loop(backend, [], "system", [], "TestBot")

        call_args = mock_exec.call_args[0]
        assert call_args[1]["amount"] == 10  # $0.01 = 10 sats

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100_000.0)
    def test_confirm_shows_converted_amount(self, mock_rate):
        """After pre-conversion, the confirm prompt shows sats (not USD)."""
        # Simulate post-pre-conversion state: amount_usd=5000 became amount=5_000_000
        desc = _describe_tool_call("fund", {
            "bot_name": "bot-1", "amount": 5_000_000,
        })
        # User sees "5,000,000 sats" — clearly wrong for "fund bot-1 with 5000 sats"
        assert "5,000,000" in desc


# ===================================================================
# Test: hallucinated tool calls that bypass confirmation
# ===================================================================

class TestHallucinatedReadOnlyTools:
    """AI might hallucinate parameters for read-only tools too.

    These don't require confirmation, so they auto-execute.
    Verify they don't cause harmful side effects.
    """

    def test_balance_check_wrong_bot_still_safe(self):
        """AI checks balance for wrong bot — harmless read-only operation."""
        # wallet_balance with a specific bot_name still only reads data
        with patch("iconfucius.cli.balance.run_all_balances",
                    return_value=None):
            with patch("iconfucius.config.require_wallet", return_value=True):
                with patch("iconfucius.config.get_bot_names",
                            return_value=["bot-1"]):
                    result = execute_tool("wallet_balance",
                                          {"bot_name": "bot-99"})
        # Returns error but no funds moved
        assert result["status"] == "error"

    def test_token_lookup_with_hallucinated_query(self):
        """AI sends random token query — harmless, just returns no results."""
        with patch("iconfucius.tokens._search_api", return_value=[]):
            result = execute_tool("token_lookup",
                                  {"query": "HALLUCINATED_TOKEN_XYZ_999"})
        assert result["status"] == "ok"
        assert result.get("api_results", []) == []



"""Tests for iconfucius.ai — OpenAICompatBackend and create_backend()."""

from unittest.mock import MagicMock, patch

import pytest

import iconfucius.config as cfg
from iconfucius.ai import (
    LlamaCppBackend, OpenAICompatBackend, cached_messages, create_backend,
)
from iconfucius.openai_compat import OpenAICompatResponse


# ---------------------------------------------------------------------------
# create_backend()
# ---------------------------------------------------------------------------

class TestCreateBackend:

    def test_create_backend_openai(self, tmp_path, monkeypatch):
        """create_backend returns OpenAICompatBackend for openai persona."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        cfg._cached_config = None
        cfg._cached_config_path = None

        persona = MagicMock()
        persona.ai_api_type = "openai"
        persona.ai_model = "test-model"
        persona.ai_base_url = "http://localhost:9999"

        backend = create_backend(persona)
        assert isinstance(backend, OpenAICompatBackend)
        assert backend.model == "test-model"
        assert backend.base_url == "http://localhost:9999"

        cfg._cached_config = None
        cfg._cached_config_path = None

    def test_legacy_alias(self):
        """LlamaCppBackend is an alias for OpenAICompatBackend."""
        assert LlamaCppBackend is OpenAICompatBackend

    def test_create_backend_unsupported(self):
        """Raise ValueError for an unrecognized ai_api_type."""
        persona = MagicMock()
        persona.ai_api_type = "unknown"
        with pytest.raises(ValueError, match="Unsupported AI API type"):
            create_backend(persona)


# ---------------------------------------------------------------------------
# LlamaCppBackend
# ---------------------------------------------------------------------------

class TestOpenAICompatBackendChat:

    def test_chat_basic(self):
        """chat() sends correct payload and returns text."""
        backend = OpenAICompatBackend(model="test-model",
                                  base_url="http://localhost:9999")
        mock_resp = MagicMock()
        mock_resp.json.return_value = {
            "choices": [
                {"message": {"role": "assistant", "content": "Hello!"}}
            ]
        }
        mock_resp.raise_for_status = MagicMock()

        with patch.object(backend._requests, "post",
                          return_value=mock_resp) as mock_post:
            result = backend.chat(
                [{"role": "user", "content": "Hi"}],
                "You are helpful.",
            )

        assert result == "Hello!"
        mock_post.assert_called_once()
        call_args = mock_post.call_args
        assert call_args[0][0] == "http://localhost:9999/v1/chat/completions"
        payload = call_args[1]["json"]
        assert payload["model"] == "test-model"
        # System prompt should be first message
        assert payload["messages"][0]["role"] == "system"
        assert payload["messages"][1]["role"] == "user"

    def test_chat_with_tools(self):
        """chat_with_tools() sends tools and returns OpenAICompatResponse."""
        backend = OpenAICompatBackend(model="test-model",
                                  base_url="http://localhost:9999")
        mock_resp = MagicMock()
        mock_resp.json.return_value = {
            "choices": [
                {
                    "message": {
                        "role": "assistant",
                        "content": None,
                        "tool_calls": [
                            {
                                "id": "call_abc",
                                "type": "function",
                                "function": {
                                    "name": "get_price",
                                    "arguments": '{"token": "BTC"}',
                                },
                            },
                        ],
                    }
                }
            ]
        }
        mock_resp.raise_for_status = MagicMock()

        tools = [
            {
                "name": "get_price",
                "description": "Get price",
                "input_schema": {
                    "type": "object",
                    "properties": {"token": {"type": "string"}},
                },
            },
        ]

        with patch.object(backend._requests, "post",
                          return_value=mock_resp) as mock_post:
            result = backend.chat_with_tools(
                [{"role": "user", "content": "price?"}],
                "sys",
                tools,
            )

        assert isinstance(result, OpenAICompatResponse)
        assert len(result.content) == 1
        block = result.content[0]
        assert block.type == "tool_use"
        assert block.name == "get_price"
        assert block.input == {"token": "BTC"}

        # Verify tools were sent in payload
        payload = mock_post.call_args[1]["json"]
        assert "tools" in payload
        assert payload["tools"][0]["type"] == "function"

    def test_list_models(self):
        """list_models() parses /v1/models response."""
        backend = OpenAICompatBackend(model="x",
                                  base_url="http://localhost:9999")
        mock_resp = MagicMock()
        mock_resp.json.return_value = {
            "data": [
                {"id": "model-a"},
                {"id": "model-b"},
            ]
        }
        mock_resp.raise_for_status = MagicMock()

        with patch.object(backend._requests, "get",
                          return_value=mock_resp):
            result = backend.list_models()

        assert result == [("model-a", "model-a"), ("model-b", "model-b")]

    def test_connection_error(self):
        """Meaningful error when server is unreachable."""
        import requests

        backend = OpenAICompatBackend(model="x",
                                  base_url="http://localhost:1")

        with patch.object(
            backend._requests, "post",
            side_effect=requests.exceptions.ConnectionError("unreachable"),
        ):
            with pytest.raises(requests.exceptions.ConnectionError):
                backend.chat([{"role": "user", "content": "hi"}], "sys")

    def test_list_models_connection_error(self):
        """list_models returns [] on connection failure."""
        backend = OpenAICompatBackend(model="x",
                                  base_url="http://localhost:1")
        assert backend.list_models() == []

    def test_chat_with_tools_caches_messages(self):
        """Multi-turn messages are sent with cache markers stripped."""
        backend = OpenAICompatBackend(model="test-model",
                                  base_url="http://localhost:9999")
        mock_resp = MagicMock()
        mock_resp.json.return_value = {
            "choices": [
                {"message": {"role": "assistant", "content": "Got it."}}
            ]
        }
        mock_resp.raise_for_status = MagicMock()

        messages = [
            {"role": "user", "content": "first"},
            {"role": "assistant", "content": "reply"},
            {"role": "user", "content": "second"},
        ]

        with patch.object(backend._requests, "post",
                          return_value=mock_resp) as mock_post:
            backend.chat_with_tools(messages, "sys", [])

        payload = mock_post.call_args[1]["json"]
        oai_msgs = payload["messages"]
        # system + 3 conversation messages
        assert len(oai_msgs) == 4
        # No cache_control keys should appear in OpenAI format
        for m in oai_msgs:
            assert "cache_control" not in m
            if isinstance(m.get("content"), list):
                for block in m["content"]:
                    assert "cache_control" not in block

    def test_base_url_trailing_slash_stripped(self):
        """Trailing slash is removed from the base URL."""
        backend = OpenAICompatBackend(model="x",
                                  base_url="http://localhost:9999/")
        assert backend.base_url == "http://localhost:9999"

    def test_model_dump_compatibility(self):
        """Verify response works with LoggingBackend's model_dump call."""
        backend = OpenAICompatBackend(model="test-model",
                                  base_url="http://localhost:9999")
        mock_resp = MagicMock()
        mock_resp.json.return_value = {
            "choices": [
                {"message": {"role": "assistant", "content": "Hi"}}
            ]
        }
        mock_resp.raise_for_status = MagicMock()

        with patch.object(backend._requests, "post",
                          return_value=mock_resp):
            result = backend.chat_with_tools(
                [{"role": "user", "content": "x"}], "sys", [],
            )

        dumped = result.model_dump(mode="json")
        assert isinstance(dumped, dict)
        assert "content" in dumped



# ---------------------------------------------------------------------------
# OpenAICompatBackend api_key support
# ---------------------------------------------------------------------------

class TestOpenAICompatBackendApiKey:

    def test_api_key_passed_to_headers(self):
        """API key is included in request headers when provided."""
        backend = OpenAICompatBackend(model="x",
                                      base_url="http://localhost:9999",
                                      api_key="sk-test-123")
        headers = backend._headers()
        assert headers == {"Authorization": "Bearer sk-test-123"}

    def test_no_api_key_empty_headers(self, monkeypatch):
        """No API key results in empty headers."""
        monkeypatch.delenv("OPENAI_API_KEY", raising=False)
        backend = OpenAICompatBackend(model="x",
                                      base_url="http://localhost:9999")
        headers = backend._headers()
        assert headers == {}

    def test_api_key_from_env(self, monkeypatch):
        """API key falls back to OPENAI_API_KEY env var."""
        monkeypatch.setenv("OPENAI_API_KEY", "sk-env-key")
        backend = OpenAICompatBackend(model="x",
                                      base_url="http://localhost:9999")
        headers = backend._headers()
        assert headers == {"Authorization": "Bearer sk-env-key"}


# ---------------------------------------------------------------------------
# cached_messages()
# ---------------------------------------------------------------------------

class TestCachedMessages:

    def test_empty_returns_empty(self):
        """Return an empty list when given no messages."""
        assert cached_messages([]) == []

    def test_single_message_unchanged(self):
        """One message — nothing to cache (no prior history)."""
        msgs = [{"role": "user", "content": "hello"}]
        result = cached_messages(msgs)
        assert result == msgs

    def test_string_content_wrapped(self):
        """String content on penultimate message is wrapped in a block."""
        msgs = [
            {"role": "user", "content": "first question"},
            {"role": "assistant", "content": "first answer"},
            {"role": "user", "content": "second question"},
        ]
        result = cached_messages(msgs)

        # Penultimate (assistant) should now have a list content block
        penultimate = result[1]
        assert isinstance(penultimate["content"], list)
        assert len(penultimate["content"]) == 1
        block = penultimate["content"][0]
        assert block["type"] == "text"
        assert block["text"] == "first answer"
        assert block["cache_control"] == {"type": "ephemeral"}

        # Other messages unchanged
        assert result[0] == msgs[0]
        assert result[2] == msgs[2]

    def test_list_content_cache_on_last_block(self):
        """List content gets cache_control on its last block."""
        msgs = [
            {"role": "user", "content": "hi"},
            {"role": "user", "content": [
                {"type": "tool_result", "tool_use_id": "t1", "content": "r1"},
                {"type": "tool_result", "tool_use_id": "t2", "content": "r2"},
            ]},
            {"role": "assistant", "content": "done"},
        ]
        result = cached_messages(msgs)

        penultimate = result[1]
        assert isinstance(penultimate["content"], list)
        # First block untouched
        assert "cache_control" not in penultimate["content"][0]
        # Last block has cache_control
        assert penultimate["content"][1]["cache_control"] == {"type": "ephemeral"}

    def test_assistant_tool_use_blocks(self):
        """Assistant message with tool_use blocks gets cache_control on last."""
        msgs = [
            {"role": "user", "content": "check balance"},
            {"role": "assistant", "content": [
                {"type": "text", "text": "Let me check..."},
                {"type": "tool_use", "id": "t1", "name": "wallet_balance",
                 "input": {}},
            ]},
            {"role": "user", "content": [
                {"type": "tool_result", "tool_use_id": "t1",
                 "content": '{"status":"ok"}'},
            ]},
        ]
        result = cached_messages(msgs)

        # Penultimate is the assistant message
        penultimate = result[1]
        blocks = penultimate["content"]
        assert "cache_control" not in blocks[0]
        assert blocks[1]["cache_control"] == {"type": "ephemeral"}

    def test_original_messages_not_mutated(self):
        """cached_messages must not mutate the original list or dicts."""
        inner = {"type": "tool_result", "tool_use_id": "t1", "content": "r"}
        msgs = [
            {"role": "user", "content": "hi"},
            {"role": "user", "content": [inner]},
            {"role": "assistant", "content": "ok"},
        ]
        _ = cached_messages(msgs)  # return value not needed; testing mutation only

        # Original inner dict should not have cache_control
        assert "cache_control" not in inner
        # Original messages list should be unchanged
        assert msgs[1]["content"][0] is inner

    def test_two_messages(self):
        """With exactly two messages, first gets cache breakpoint."""
        msgs = [
            {"role": "user", "content": "question"},
            {"role": "assistant", "content": "answer"},
        ]
        result = cached_messages(msgs)

        first = result[0]
        assert isinstance(first["content"], list)
        assert first["content"][0]["cache_control"] == {"type": "ephemeral"}
        # Second unchanged
        assert result[1] == msgs[1]



"""Tests for iconfucius.skills.executor — Tool dispatch and execution."""

import os
from unittest.mock import MagicMock, patch

from iconfucius.skills.executor import (
    execute_tool,
    _enable_verify_certificates,
    _resolve_bot_names,
    _tokens_to_subunits,
    _usd_to_sats,
    _usd_to_tokens,
)

_FAKE_PEM = "fake.pem"


class TestResolveBotNames:
    def test_single_bot_name(self):
        """Verify single bot name."""
        assert _resolve_bot_names({"bot_name": "bot-1"}) == ["bot-1"]

    def test_bot_names_list(self):
        """Verify bot names list."""
        result = _resolve_bot_names({"bot_names": ["bot-12", "bot-14"]})
        assert result == ["bot-12", "bot-14"]

    def test_all_bots(self, odin_project):  # noqa: ARG002
        """Verify all bots."""
        result = _resolve_bot_names({"all_bots": True})
        assert len(result) == 3  # default odin_project fixture has 3 bots

    def test_empty_returns_empty(self):
        """Verify empty returns empty."""
        assert _resolve_bot_names({}) == []

    def test_bot_names_takes_priority_over_bot_name(self):
        """Verify bot names takes priority over bot name."""
        result = _resolve_bot_names({
            "bot_names": ["bot-2", "bot-3"],
            "bot_name": "bot-1",
        })
        assert result == ["bot-2", "bot-3"]

    def test_all_bots_takes_priority(self, odin_project):  # noqa: ARG002
        """Verify all bots takes priority."""
        result = _resolve_bot_names({
            "all_bots": True,
            "bot_name": "bot-1",
            "bot_names": ["bot-2"],
        })
        assert len(result) == 3  # all configured bots, not the explicit ones


class TestExecuteToolDispatch:
    def test_unknown_tool_returns_error(self):
        """Verify unknown tool returns error."""
        result = execute_tool("nonexistent_tool", {})
        assert result["status"] == "error"
        assert "Unknown tool" in result["error"]



class TestSetupStatusExecutor:
    def test_setup_status_returns_all_fields(self):
        """Verify setup status returns all fields."""
        result = execute_tool("setup_status", {})
        assert result["status"] == "ok"
        assert "config_exists" in result
        assert "wallet_exists" in result
        assert "env_exists" in result
        assert "has_api_key" in result
        assert "ready" in result

    def test_setup_status_ready_requires_all(self):
        """ready should be False when any component is missing."""
        # In the test environment, wallet likely doesn't exist
        result = execute_tool("setup_status", {})
        assert result["status"] == "ok"
        # ready should be a bool
        assert isinstance(result["ready"], bool)


class TestInitExecutor:
    def test_init_creates_config(self, tmp_path, monkeypatch):
        """Verify init creates config."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        # Clear config cache
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None

        result = execute_tool("init", {})
        assert result["status"] == "ok"
        assert (tmp_path / "iconfucius.toml").exists()

    def test_init_existing_config_returns_error(self, tmp_path, monkeypatch):
        """Verify init existing config returns error."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        (tmp_path / "iconfucius.toml").write_text("[settings]\n")

        result = execute_tool("init", {})
        assert result["status"] == "error"
        assert "already exists" in result["error"]

    def test_init_with_num_bots(self, tmp_path, monkeypatch):
        """Verify init with num bots."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None

        result = execute_tool("init", {"num_bots": 2})
        assert result["status"] == "ok"
        content = (tmp_path / "iconfucius.toml").read_text()
        assert "[bots.bot-1]" in content
        assert "[bots.bot-2]" in content
        assert "[bots.bot-3]" not in content

    def test_init_without_num_bots_defaults_to_three(self, tmp_path, monkeypatch):
        """Verify init without num bots defaults to three."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None

        result = execute_tool("init", {})
        assert result["status"] == "ok"
        content = (tmp_path / "iconfucius.toml").read_text()
        assert "[bots.bot-3]" in content
        assert "[bots.bot-4]" not in content


class TestBotListExecutor:
    """Tests for bot_list agent skill."""

    def test_lists_bots(self, tmp_path, monkeypatch):
        """Verify lists bots."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        execute_tool("init", {"num_bots": 5})
        cfg._cached_config = None
        cfg._cached_config_path = None

        result = execute_tool("bot_list", {})
        assert result["status"] == "ok"
        assert result["bot_count"] == 5
        assert result["bot_names"] == ["bot-1", "bot-2", "bot-3", "bot-4", "bot-5"]
        assert "5 bot(s)" in result["display"]

    def test_no_config_returns_error(self, tmp_path, monkeypatch):
        """Verify no config returns error."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None

        result = execute_tool("bot_list", {})
        assert result["status"] == "error"


class TestSetBotCountExecutor:
    """Tests for set_bot_count agent skill."""

    def _setup_project(self, tmp_path, monkeypatch, num_bots=3):
        """Helper: init a project with N bots in tmp_path."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        result = execute_tool("init", {"num_bots": num_bots})
        assert result["status"] == "ok"
        cfg._cached_config = None
        cfg._cached_config_path = None
        return tmp_path

    def test_no_config_returns_error(self, tmp_path, monkeypatch):
        """Verify no config returns error."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        result = execute_tool("set_bot_count", {"num_bots": 5})
        assert result["status"] == "error"
        assert "No iconfucius.toml" in result["error"]

    def test_same_count_is_noop(self, tmp_path, monkeypatch):
        """Verify same count is noop."""
        self._setup_project(tmp_path, monkeypatch, num_bots=3)
        result = execute_tool("set_bot_count", {"num_bots": 3})
        assert result["status"] == "ok"
        assert result["bot_count"] == 3
        assert "Already" in result["message"]

    def test_increase_adds_bots(self, tmp_path, monkeypatch):
        """Verify increase adds bots."""
        self._setup_project(tmp_path, monkeypatch, num_bots=3)
        result = execute_tool("set_bot_count", {"num_bots": 7})
        assert result["status"] == "ok"
        assert result["bot_count"] == 7
        assert len(result["bots_added"]) == 4
        content = (tmp_path / "iconfucius.toml").read_text()
        for i in range(1, 8):
            assert f"[bots.bot-{i}]" in content
        assert "[bots.bot-8]" not in content

    def test_increase_large(self, tmp_path, monkeypatch):
        """Verify increase large."""
        self._setup_project(tmp_path, monkeypatch, num_bots=3)
        result = execute_tool("set_bot_count", {"num_bots": 100})
        assert result["status"] == "ok"
        assert result["bot_count"] == 100
        assert len(result["bots_added"]) == 97

    def test_decrease_no_sessions_removes_immediately(self, tmp_path, monkeypatch):
        """Bots without cached sessions are removed without balance check."""
        self._setup_project(tmp_path, monkeypatch, num_bots=5)
        result = execute_tool("set_bot_count", {"num_bots": 2})
        assert result["status"] == "ok"
        assert result["bot_count"] == 2
        assert set(result["bots_removed"]) == {"bot-3", "bot-4", "bot-5"}
        content = (tmp_path / "iconfucius.toml").read_text()
        assert "[bots.bot-1]" in content
        assert "[bots.bot-2]" in content
        assert "[bots.bot-3]" not in content

    def test_decrease_with_holdings_returns_blocked(self, tmp_path, monkeypatch):
        """Bots with cached sessions and holdings block removal."""
        self._setup_project(tmp_path, monkeypatch, num_bots=3)
        # Create a fake cached session for bot-3
        cache_dir = tmp_path / ".cache"
        cache_dir.mkdir(exist_ok=True)
        (cache_dir / "session_bot-3.json").write_text("{}")

        from iconfucius.cli.balance import BotBalances

        fake_data = BotBalances(
            bot_name="bot-3", bot_principal="abc-123",
            odin_sats=5000, token_holdings=[{"ticker": "TEST", "balance": 100}],
        )
        with patch("iconfucius.cli.balance.collect_balances", return_value=fake_data):
            result = execute_tool("set_bot_count", {"num_bots": 2})
        assert result["status"] == "blocked"
        assert result["reason"] == "bots_have_holdings"
        assert len(result["holdings"]) == 1
        assert result["holdings"][0]["bot_name"] == "bot-3"
        # Config should NOT have been modified
        content = (tmp_path / "iconfucius.toml").read_text()
        assert "[bots.bot-3]" in content

    def test_decrease_force_skips_check(self, tmp_path, monkeypatch):
        """force=True removes bots without checking holdings."""
        self._setup_project(tmp_path, monkeypatch, num_bots=5)
        # Create fake sessions (would trigger balance check without force)
        cache_dir = tmp_path / ".cache"
        cache_dir.mkdir(exist_ok=True)
        (cache_dir / "session_bot-4.json").write_text("{}")
        (cache_dir / "session_bot-5.json").write_text("{}")

        result = execute_tool("set_bot_count", {"num_bots": 3, "force": True})
        assert result["status"] == "ok"
        assert result["bot_count"] == 3
        content = (tmp_path / "iconfucius.toml").read_text()
        assert "[bots.bot-3]" in content
        assert "[bots.bot-4]" not in content

    def test_num_bots_required(self, tmp_path, monkeypatch):
        """Verify num bots required."""
        self._setup_project(tmp_path, monkeypatch, num_bots=3)
        result = execute_tool("set_bot_count", {})
        assert result["status"] == "error"
        assert "required" in result["error"].lower()

    def test_num_bots_clamped(self, tmp_path, monkeypatch):
        """num_bots is clamped to 1-1000 range."""
        self._setup_project(tmp_path, monkeypatch, num_bots=3)
        result = execute_tool("set_bot_count", {"num_bots": 0})
        # 0 clamped to 1, so decrease from 3 to 1
        assert result["status"] == "ok"
        assert result["bot_count"] == 1


class TestWalletCreateExecutor:
    def test_wallet_create_creates_pem(self, tmp_path, monkeypatch):
        """Verify wallet create creates pem."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))

        result = execute_tool("wallet_create", {})
        assert result["status"] == "ok"
        pem_path = tmp_path / ".wallet" / "identity-private.pem"
        assert pem_path.exists()

    def test_wallet_create_existing_returns_error(self, tmp_path, monkeypatch):
        """Verify wallet create existing returns error."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        wallet_dir = tmp_path / ".wallet"
        wallet_dir.mkdir()
        (wallet_dir / "identity-private.pem").write_text("existing")

        result = execute_tool("wallet_create", {})
        assert result["status"] == "error"
        assert "already exists" in result["error"]


class TestTokenLookupExecutor:
    def test_token_lookup_known_token(self):
        """token_lookup should find IConfucius by name."""
        with patch("iconfucius.tokens._search_api", return_value=[]):
            result = execute_tool("token_lookup", {"query": "IConfucius"})
        assert result["status"] == "ok"
        assert result["known_match"] is not None
        assert result["known_match"]["id"] == "29m8"

    def test_token_lookup_missing_query_returns_error(self):
        """Verify token lookup missing query returns error."""
        result = execute_tool("token_lookup", {})
        assert result["status"] == "error"
        assert "required" in result["error"].lower()


class TestSecurityStatusExecutor:
    """Tests for security_status agent skill."""

    def _setup_project(self, tmp_path, monkeypatch, settings=""):
        """Set up a test project directory."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        content = '[settings]\n' + settings + '\n[bots.bot-1]\ndescription = "Bot 1"\n'
        (tmp_path / "iconfucius.toml").write_text(content)

    def test_blst_not_installed(self, tmp_path, monkeypatch):
        """Verify blst not installed."""
        self._setup_project(tmp_path, monkeypatch)
        with patch.dict("sys.modules", {"blst": None}):
            result = execute_tool("security_status", {})
        assert result["status"] == "ok"
        assert result["blst_installed"] is False
        assert result["verify_certificates"] is False
        assert "not installed" in result["display"]

    def test_blst_installed_not_enabled(self, tmp_path, monkeypatch):
        """Verify blst installed not enabled."""
        self._setup_project(tmp_path, monkeypatch)
        with patch.dict("sys.modules", {"blst": object()}):
            result = execute_tool("security_status", {})
        assert result["status"] == "ok"
        assert result["blst_installed"] is True
        assert result["verify_certificates"] is False
        assert "disabled" in result["display"]
        assert "enable" in result["display"].lower()

    def test_blst_installed_and_enabled(self, tmp_path, monkeypatch):
        """Verify blst installed and enabled."""
        self._setup_project(tmp_path, monkeypatch,
                            settings="verify_certificates = true")
        with patch.dict("sys.modules", {"blst": object()}):
            result = execute_tool("security_status", {})
        assert result["status"] == "ok"
        assert result["blst_installed"] is True
        assert result["verify_certificates"] is True
        assert "enabled" in result["display"]

    def test_cache_sessions_disabled(self, tmp_path, monkeypatch):
        """Verify cache sessions disabled."""
        self._setup_project(tmp_path, monkeypatch,
                            settings="cache_sessions = false")
        with patch.dict("sys.modules", {"blst": None}):
            result = execute_tool("security_status", {})
        assert result["cache_sessions"] is False
        assert "disabled" in result["display"].lower()

    def test_recommendations_when_blst_missing(self, tmp_path, monkeypatch):
        """Verify recommendations when blst missing."""
        self._setup_project(tmp_path, monkeypatch)
        with patch.dict("sys.modules", {"blst": None}):
            result = execute_tool("security_status", {})
        assert "Recommendations:" in result["display"]
        assert "install_blst" in result["display"].lower()

    def test_recommendation_when_blst_present_but_not_enabled(
        self, tmp_path, monkeypatch
    ):
        """Verify recommendation when blst present but not enabled."""
        self._setup_project(tmp_path, monkeypatch)
        with patch.dict("sys.modules", {"blst": object()}):
            result = execute_tool("security_status", {})
        assert "Recommendations:" in result["display"]
        assert "verify_certificates" in result["display"]

    def test_no_recommendations_when_fully_configured(
        self, tmp_path, monkeypatch
    ):
        """Verify no recommendations when fully configured."""
        self._setup_project(tmp_path, monkeypatch,
                            settings="verify_certificates = true")
        with patch.dict("sys.modules", {"blst": object()}):
            result = execute_tool("security_status", {})
        assert "Recommendations:" not in result["display"]


class TestEnableVerifyCertificates:
    """Tests for _enable_verify_certificates helper."""

    def test_no_config_returns_not_enabled(self, tmp_path, monkeypatch):
        """Verify no config returns not enabled."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        result = _enable_verify_certificates()
        assert result["enabled_now"] is False

    def test_enables_when_not_present(self, tmp_path, monkeypatch):
        """Verify enables when not present."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        (tmp_path / "iconfucius.toml").write_text(
            '[settings]\n[bots.bot-1]\ndescription = "Bot 1"\n'
        )
        result = _enable_verify_certificates()
        assert result["enabled_now"] is True
        content = (tmp_path / "iconfucius.toml").read_text()
        assert "verify_certificates = true" in content

    def test_enables_when_false(self, tmp_path, monkeypatch):
        """Verify enables when false."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        (tmp_path / "iconfucius.toml").write_text(
            '[settings]\nverify_certificates = false\n'
            '[bots.bot-1]\ndescription = "Bot 1"\n'
        )
        result = _enable_verify_certificates()
        assert result["enabled_now"] is True
        content = (tmp_path / "iconfucius.toml").read_text()
        assert "verify_certificates = true" in content
        assert "verify_certificates = false" not in content

    def test_noop_when_already_true(self, tmp_path, monkeypatch):
        """Verify noop when already true."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        (tmp_path / "iconfucius.toml").write_text(
            '[settings]\nverify_certificates = true\n'
            '[bots.bot-1]\ndescription = "Bot 1"\n'
        )
        result = _enable_verify_certificates()
        assert result["enabled_now"] is False

    def test_adds_settings_section_if_missing(self, tmp_path, monkeypatch):
        """Verify adds settings section if missing."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        (tmp_path / "iconfucius.toml").write_text(
            '[bots.bot-1]\ndescription = "Bot 1"\n'
        )
        result = _enable_verify_certificates()
        assert result["enabled_now"] is True
        content = (tmp_path / "iconfucius.toml").read_text()
        assert "verify_certificates = true" in content


class TestInstallBlstExecutor:
    """Tests for install_blst agent skill."""

    def test_already_installed_enables_config(self, tmp_path, monkeypatch):
        """When blst is already importable, enables verify_certificates."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        (tmp_path / "iconfucius.toml").write_text(
            '[settings]\n[bots.bot-1]\ndescription = "Bot 1"\n'
        )
        with patch.dict("sys.modules", {"blst": object()}):
            result = execute_tool("install_blst", {})
        assert result["status"] == "ok"
        assert "already installed" in result["display"]
        assert "Enabled" in result["display"]
        content = (tmp_path / "iconfucius.toml").read_text()
        assert "verify_certificates = true" in content

    def test_already_installed_already_enabled(self, tmp_path, monkeypatch):
        """When blst installed and verify_certificates already true."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        (tmp_path / "iconfucius.toml").write_text(
            '[settings]\nverify_certificates = true\n'
            '[bots.bot-1]\ndescription = "Bot 1"\n'
        )
        with patch.dict("sys.modules", {"blst": object()}):
            result = execute_tool("install_blst", {})
        assert result["status"] == "ok"
        assert "already installed" in result["display"]
        assert "already enabled" in result["display"]

    def test_missing_prerequisites(self, tmp_path, monkeypatch):
        """Reports missing tools when blst not installed."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None
        with patch.dict("sys.modules", {"blst": None}):
            with patch("shutil.which", return_value=None):
                result = execute_tool("install_blst", {})
        assert result["status"] == "error"
        assert "Missing prerequisites" in result["error"]

    def test_missing_swig_only(self, tmp_path, monkeypatch):
        """Reports only swig missing when git and cc present."""
        monkeypatch.chdir(tmp_path)
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        import iconfucius.config as cfg
        cfg._cached_config = None
        cfg._cached_config_path = None

        def fake_which(cmd):
            """Mock shutil.which for testing."""
            if cmd in ("git", "cc"):
                return f"/usr/bin/{cmd}"
            return None

        with patch.dict("sys.modules", {"blst": None}):
            with patch("shutil.which", side_effect=fake_which):
                result = execute_tool("install_blst", {})
        assert result["status"] == "error"
        assert "swig" in result["error"]
        assert "git" not in result["error"].split("Missing")[1]


class TestTradeRecording:
    """Tests that buy/sell tool calls record trades to memory."""

    def _fake_handler(self, result):
        """Return a handler function that returns the given result."""
        return lambda args: result

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"price": 1500, "ticker": "ICONFUCIUS"})
    @patch("iconfucius.memory.append_trade")
    def test_buy_records_trade(self, mock_append, _mock_fetch, _mock_usd,
                               tmp_path, monkeypatch):
        """Verify buy records trade."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        from iconfucius.skills.executor import _HANDLERS
        original = _HANDLERS["trade_buy"]
        _HANDLERS["trade_buy"] = self._fake_handler(
            {"status": "ok", "display": "Bought!"})
        try:
            result = execute_tool("trade_buy",
                                  {"token_id": "29m8", "amount": 1000,
                                   "bot_name": "bot-1"},
                                  persona_name="iconfucius")
        finally:
            _HANDLERS["trade_buy"] = original
        assert result["status"] == "ok"
        mock_append.assert_called_once()
        entry = mock_append.call_args[0][1]
        assert entry["action"] == "BUY"
        assert entry["token_id"] == "29m8"
        assert entry["ticker"] == "ICONFUCIUS"
        assert "bot-1" in entry["bots"]
        assert entry["amount_sats"] == 1000
        assert entry["price_sats"] == 1500
        assert "est_tokens" in entry

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"price": 1500, "ticker": "ICONFUCIUS"})
    @patch("iconfucius.memory.append_trade")
    def test_sell_records_trade(self, mock_append, _mock_fetch, _mock_usd,
                                tmp_path, monkeypatch):
        """Verify sell records trade."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        from iconfucius.skills.executor import _HANDLERS
        original = _HANDLERS["trade_sell"]
        _HANDLERS["trade_sell"] = self._fake_handler(
            {"status": "ok", "display": "Sold!"})
        try:
            result = execute_tool("trade_sell",
                                  {"token_id": "29m8", "amount": 5000000,
                                   "bot_name": "bot-1"},
                                  persona_name="iconfucius")
        finally:
            _HANDLERS["trade_sell"] = original
        assert result["status"] == "ok"
        mock_append.assert_called_once()
        entry = mock_append.call_args[0][1]
        assert entry["action"] == "SELL"
        assert entry["token_id"] == "29m8"
        assert entry["ticker"] == "ICONFUCIUS"
        assert entry["price_sats"] == 1500
        assert "est_sats_received" in entry

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"price": 1500, "ticker": "ICONFUCIUS"})
    @patch("iconfucius.memory.append_trade")
    def test_sell_all_records_trade(self, mock_append, _mock_fetch, _mock_usd,
                                    tmp_path, monkeypatch):
        """Verify sell all records trade."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        from iconfucius.skills.executor import _HANDLERS
        original = _HANDLERS["trade_sell"]
        _HANDLERS["trade_sell"] = self._fake_handler(
            {"status": "ok", "display": "Sold!"})
        try:
            result = execute_tool("trade_sell",
                                  {"token_id": "29m8", "amount": "all",
                                   "all_bots": True},
                                  persona_name="iconfucius")
        finally:
            _HANDLERS["trade_sell"] = original
        assert result["status"] == "ok"
        mock_append.assert_called_once()
        entry = mock_append.call_args[0][1]
        assert entry["action"] == "SELL"
        assert entry["tokens_sold"] == "all"
        assert entry["token_id"] == "29m8"
        # sell-all should NOT have est_sats_received
        assert "est_sats_received" not in entry

    @patch("iconfucius.memory.append_trade")
    def test_failed_trade_not_recorded(self, mock_append):
        """Verify failed trade not recorded."""
        from iconfucius.skills.executor import _HANDLERS
        original = _HANDLERS["trade_buy"]
        _HANDLERS["trade_buy"] = self._fake_handler(
            {"status": "error", "error": "No wallet"})
        try:
            result = execute_tool("trade_buy",
                                  {"token_id": "29m8", "amount": 1000,
                                   "bot_name": "bot-1"},
                                  persona_name="iconfucius")
        finally:
            _HANDLERS["trade_buy"] = original
        assert result["status"] == "error"
        mock_append.assert_not_called()

    @patch("iconfucius.memory.append_trade")
    def test_no_persona_no_recording(self, mock_append):
        """Verify no persona no recording."""
        from iconfucius.skills.executor import _HANDLERS
        original = _HANDLERS["trade_buy"]
        _HANDLERS["trade_buy"] = self._fake_handler(
            {"status": "ok", "display": "Bought!"})
        try:
            result = execute_tool("trade_buy",
                                  {"token_id": "29m8", "amount": 1000,
                                   "bot_name": "bot-1"})
        finally:
            _HANDLERS["trade_buy"] = original
        assert result["status"] == "ok"
        mock_append.assert_not_called()

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"price": 1500, "ticker": "ICONFUCIUS"})
    @patch("iconfucius.memory.append_trade",
           side_effect=Exception("disk full"))
    def test_recording_failure_is_silent(self, mock_append, _mock_fetch,
                                         _mock_usd):
        """Trade recording errors don't break the trade result."""
        from iconfucius.skills.executor import _HANDLERS
        original = _HANDLERS["trade_buy"]
        _HANDLERS["trade_buy"] = self._fake_handler(
            {"status": "ok", "display": "Bought!"})
        try:
            result = execute_tool("trade_buy",
                                  {"token_id": "29m8", "amount": 1000,
                                   "bot_name": "bot-1"},
                                  persona_name="iconfucius")
        finally:
            _HANDLERS["trade_buy"] = original
        assert result["status"] == "ok"


class TestMemoryToolHandlers:
    """Tests for memory_read_strategy, memory_read_learnings, memory_update."""

    def test_read_strategy_empty(self, tmp_path, monkeypatch):
        """Verify read strategy empty."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        result = execute_tool("memory_read_strategy", {},
                              persona_name="test-persona")
        assert result["status"] == "ok"
        assert "No strategy" in result["display"]

    def test_read_strategy_with_content(self, tmp_path, monkeypatch):
        """Verify read strategy with content."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        from iconfucius.memory import write_strategy
        write_strategy("test-persona", "# My Strategy\nBuy low sell high")
        result = execute_tool("memory_read_strategy", {},
                              persona_name="test-persona")
        assert result["status"] == "ok"
        assert "Buy low sell high" in result["display"]

    def test_read_learnings_empty(self, tmp_path, monkeypatch):
        """Verify read learnings empty."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        result = execute_tool("memory_read_learnings", {},
                              persona_name="test-persona")
        assert result["status"] == "ok"
        assert "No learnings" in result["display"]

    def test_read_learnings_with_content(self, tmp_path, monkeypatch):
        """Verify read learnings with content."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        from iconfucius.memory import write_learnings
        write_learnings("test-persona", "# Learnings\nVolume spikes matter")
        result = execute_tool("memory_read_learnings", {},
                              persona_name="test-persona")
        assert result["status"] == "ok"
        assert "Volume spikes matter" in result["display"]

    def test_update_strategy(self, tmp_path, monkeypatch):
        """Verify update strategy."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        result = execute_tool("memory_update",
                              {"file": "strategy", "content": "New strategy"},
                              persona_name="test-persona")
        assert result["status"] == "ok"
        assert "updated" in result["display"].lower()
        from iconfucius.memory import read_strategy
        assert read_strategy("test-persona") == "New strategy"

    def test_update_learnings(self, tmp_path, monkeypatch):
        """Verify update learnings."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        result = execute_tool("memory_update",
                              {"file": "learnings", "content": "New learnings"},
                              persona_name="test-persona")
        assert result["status"] == "ok"
        from iconfucius.memory import read_learnings
        assert read_learnings("test-persona") == "New learnings"

    def test_update_invalid_file(self, tmp_path, monkeypatch):
        """Verify update invalid file."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        result = execute_tool("memory_update",
                              {"file": "trades", "content": "Nope"},
                              persona_name="test-persona")
        assert result["status"] == "error"
        assert "Unknown file" in result["error"]

    def test_update_missing_params(self, tmp_path, monkeypatch):
        """Verify update missing params."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        result = execute_tool("memory_update", {},
                              persona_name="test-persona")
        assert result["status"] == "error"
        assert "required" in result["error"].lower()

    def test_no_persona_returns_error(self):
        """Verify no persona returns error."""
        result = execute_tool("memory_read_strategy", {})
        assert result["status"] == "error"
        assert "persona" in result["error"].lower()


class TestWalletBalanceResult:
    """wallet_balance returns structured JSON for the AI to summarize."""

    def test_returns_structured_data(self):
        """Verify returns structured data."""
        fake_data = {
            "wallet_ckbtc_sats": 1000,
            "bots": [
                {"name": "bot-1", "principal": "bot1-principal-abc",
                 "odin_sats": 500, "has_odin_account": True,
                 "tokens": [
                    {"ticker": "CONF", "id": "29m8", "balance": 0.000001,
                     "div": 8, "value_sats": 200},
                ]},
            ],
            "totals": {
                "odin_sats": 500,
                "token_value_sats": 200,
                "portfolio_sats": 1700,
            },
            "_display": "table output",
        }
        with patch("iconfucius.cli.balance.run_all_balances",
                    return_value=fake_data):
            with patch("iconfucius.config.require_wallet", return_value=True):
                with patch("iconfucius.config.get_bot_names",
                            return_value=["bot-1"]):
                    result = execute_tool("wallet_balance", {})
        assert result["status"] == "ok"
        assert result["wallet_ckbtc_sats"] == 1000
        assert result["total_odin_sats"] == 500
        assert result["total_token_value_sats"] == 200
        assert result["portfolio_sats"] == 1700
        # Per-bot data included so AI doesn't need individual balance calls
        assert "bots" in result
        assert len(result["bots"]) == 1
        assert result["bots"][0]["name"] == "bot-1"
        assert result["bots"][0]["principal"] == "bot1-principal-abc"
        assert result["bots"][0]["odin_sats"] == 500
        assert result["bots"][0]["has_odin_account"] is True
        assert result["bots"][0]["tokens"][0]["ticker"] == "CONF"
        # No terminal output — AI summarizes from structured data
        assert "_terminal_output" not in result
        # Constraints surfaced for AI decision-making
        assert result["constraints"]["min_deposit_sats"] == 5000
        assert result["constraints"]["min_trade_sats"] == 500

    def test_includes_constraints(self):
        """constraints dict uses config constants so the AI respects minimums."""
        fake_data = {
            "wallet_ckbtc_sats": 0,
            "totals": {"odin_sats": 0, "token_value_sats": 0, "portfolio_sats": 0},
            "_display": "",
        }
        with patch("iconfucius.cli.balance.run_all_balances",
                    return_value=fake_data):
            with patch("iconfucius.config.require_wallet", return_value=True):
                with patch("iconfucius.config.get_bot_names",
                            return_value=["bot-1"]):
                    result = execute_tool("wallet_balance", {})
        assert result["status"] == "ok"
        assert result["constraints"] == {
            "min_deposit_sats": 5000,
            "min_trade_sats": 500,
        }

    def test_none_data_returns_error_with_funding_hint(self):
        """None from run_all_balances returns error with funding hint."""
        with patch("iconfucius.cli.balance.run_all_balances",
                    return_value=None):
            with patch("iconfucius.config.require_wallet", return_value=True):
                with patch("iconfucius.config.get_bot_names",
                            return_value=["bot-1"]):
                    result = execute_tool("wallet_balance", {})
        assert result["status"] == "error"
        assert "how_to_fund_wallet" in result["error"]
        assert "funding" in result["error"].lower() or "fund" in result["error"].lower()

    def test_bot_note_passed_through(self):
        """Bot-level note (e.g. unfunded wallet) is passed through to result."""
        fake_data = {
            "wallet_ckbtc_sats": 0,
            "totals": {"odin_sats": 0, "token_value_sats": 0, "portfolio_sats": 0},
            "bots": [{
                "name": "bot-1",
                "principal": "",
                "odin_sats": None,
                "tokens": None,
                "has_odin_account": None,
                "note": "Balance could not be checked — wallet needs funding for signing fees. Do NOT report this bot's balance as 0. Use the how_to_fund_wallet tool for instructions.",
            }],
            "_display": "",
        }
        with patch("iconfucius.cli.balance.run_all_balances",
                    return_value=fake_data):
            with patch("iconfucius.config.require_wallet", return_value=True):
                with patch("iconfucius.config.get_bot_names",
                            return_value=["bot-1"]):
                    result = execute_tool("wallet_balance", {})
        assert result["status"] == "ok"
        assert result["bots"][0]["note"] == "Balance could not be checked — wallet needs funding for signing fees. Do NOT report this bot's balance as 0. Use the how_to_fund_wallet tool for instructions."
        assert "odin_sats" not in result["bots"][0]
        assert "tokens" not in result["bots"][0]
        assert "has_odin_account" not in result["bots"][0]


    def test_next_step_wallet_empty_bots_unchecked(self):
        """Wallet empty + unchecked bots → next_step says fund wallet."""
        fake_data = {
            "wallet_ckbtc_sats": 0,
            "totals": {"odin_sats": 0, "token_value_sats": 0, "portfolio_sats": 0},
            "bots": [{
                "name": "bot-1", "principal": "",
                "odin_sats": None, "tokens": None, "has_odin_account": None,
                "note": "Balance could not be checked.",
            }],
            "_display": "",
        }
        with patch("iconfucius.cli.balance.run_all_balances",
                    return_value=fake_data):
            with patch("iconfucius.config.require_wallet", return_value=True):
                with patch("iconfucius.config.get_bot_names",
                            return_value=["bot-1"]):
                    result = execute_tool("wallet_balance", {})
        assert "next_step" in result
        assert "how_to_fund_wallet" in result["next_step"]

    def test_next_step_wallet_empty_bots_zero(self):
        """Wallet empty + all bot balances zero → next_step says fund wallet."""
        fake_data = {
            "wallet_ckbtc_sats": 0,
            "totals": {"odin_sats": 0, "token_value_sats": 0, "portfolio_sats": 0},
            "bots": [{
                "name": "bot-1", "principal": "p1",
                "odin_sats": 0, "tokens": [], "has_odin_account": True,
            }],
            "_display": "",
        }
        with patch("iconfucius.cli.balance.run_all_balances",
                    return_value=fake_data):
            with patch("iconfucius.config.require_wallet", return_value=True):
                with patch("iconfucius.config.get_bot_names",
                            return_value=["bot-1"]):
                    result = execute_tool("wallet_balance", {})
        assert "next_step" in result
        assert "how_to_fund_wallet" in result["next_step"]

    def test_next_step_wallet_empty_bots_have_holdings(self):
        """Wallet empty + bots have holdings → next_step mentions options."""
        fake_data = {
            "wallet_ckbtc_sats": 0,
            "totals": {"odin_sats": 5000, "token_value_sats": 200,
                       "portfolio_sats": 5200},
            "bots": [{
                "name": "bot-1", "principal": "p1",
                "odin_sats": 5000, "tokens": [
                    {"ticker": "TEST", "balance": 1, "value_sats": 200}],
                "has_odin_account": True,
            }],
            "_display": "",
        }
        with patch("iconfucius.cli.balance.run_all_balances",
                    return_value=fake_data):
            with patch("iconfucius.config.require_wallet", return_value=True):
                with patch("iconfucius.config.get_bot_names",
                            return_value=["bot-1"]):
                    result = execute_tool("wallet_balance", {})
        assert "next_step" in result
        assert "holdings" in result["next_step"]
        assert "withdraw" in result["next_step"]

    def test_no_next_step_when_wallet_funded(self):
        """Wallet with funds → no next_step field."""
        fake_data = {
            "wallet_ckbtc_sats": 10000,
            "totals": {"odin_sats": 0, "token_value_sats": 0,
                       "portfolio_sats": 10000},
            "bots": [{
                "name": "bot-1", "principal": "p1",
                "odin_sats": 0, "tokens": [], "has_odin_account": True,
            }],
            "_display": "",
        }
        with patch("iconfucius.cli.balance.run_all_balances",
                    return_value=fake_data):
            with patch("iconfucius.config.require_wallet", return_value=True):
                with patch("iconfucius.config.get_bot_names",
                            return_value=["bot-1"]):
                    result = execute_tool("wallet_balance", {})
        assert "next_step" not in result


class TestTokenDiscoverExecutor:
    """Tests for the token_discover agent skill."""

    _FAKE_TOKENS = [
        {
            "id": "abc1",
            "name": "AlphaToken",
            "ticker": "ALPHA",
            "price_sats": 2.0,
            "marketcap_sats": 50000000,
            "volume_24h_sats": 10000000,
            "holder_count": 150,
            "bonded": True,
            "twitter_verified": True,
            "safety": "bonded (graduated to AMM) · Twitter verified · 150 holders · not in known tokens registry",
        },
        {
            "id": "xyz2",
            "name": "BetaToken",
            "ticker": "BETA",
            "price_sats": 0.5,
            "marketcap_sats": 5000000,
            "volume_24h_sats": 1000000,
            "holder_count": 30,
            "bonded": True,
            "twitter_verified": False,
            "safety": "bonded (graduated to AMM) · Twitter NOT verified · 30 holders · not in known tokens registry",
        },
    ]

    def test_returns_tokens(self):
        """Verify returns tokens."""
        with patch("iconfucius.tokens.discover_tokens",
                    return_value=self._FAKE_TOKENS):
            with patch("iconfucius.config.get_btc_to_usd_rate",
                        return_value=100000.0):
                result = execute_tool("token_discover",
                                      {"sort": "volume", "limit": 10})
        assert result["status"] == "ok"
        assert result["count"] == 2
        ids = [t["id"] for t in result["tokens"]]
        assert "abc1" in ids
        assert "xyz2" in ids

    def test_default_sort_is_volume(self):
        """Verify default sort is volume."""
        with patch("iconfucius.tokens.discover_tokens",
                    return_value=self._FAKE_TOKENS):
            with patch("iconfucius.config.get_btc_to_usd_rate",
                        return_value=100000.0):
                result = execute_tool("token_discover", {})
        assert result["status"] == "ok"
        assert result["sort"] == "volume"
        assert "trending" in result["display"].lower()

    def test_newest_sort(self):
        """Verify newest sort."""
        with patch("iconfucius.tokens.discover_tokens",
                    return_value=self._FAKE_TOKENS):
            with patch("iconfucius.config.get_btc_to_usd_rate",
                        return_value=100000.0):
                result = execute_tool("token_discover", {"sort": "newest"})
        assert result["status"] == "ok"
        assert result["sort"] == "newest"
        assert "newest" in result["display"].lower()

    def test_token_fields_present(self):
        """Verify token fields present."""
        with patch("iconfucius.tokens.discover_tokens",
                    return_value=self._FAKE_TOKENS[:1]):
            with patch("iconfucius.config.get_btc_to_usd_rate",
                        return_value=100000.0):
                result = execute_tool("token_discover", {"limit": 1})
        assert result["status"] == "ok"
        token = result["tokens"][0]
        assert "id" in token
        assert "name" in token
        assert "ticker" in token
        assert "price_sats" in token
        assert "marketcap_sats" in token
        assert "volume_24h_sats" in token
        assert "holder_count" in token
        assert "bonded" in token
        assert "safety" in token

    def test_empty_results(self):
        """Verify empty results."""
        with patch("iconfucius.tokens.discover_tokens", return_value=[]):
            with patch("iconfucius.config.get_btc_to_usd_rate",
                        return_value=100000.0):
                result = execute_tool("token_discover", {})
        assert result["status"] == "ok"
        assert result["tokens"] == []
        assert result["count"] == 0
        assert "No tokens found" in result["display"]

    def test_display_includes_token_info(self):
        """Verify display includes token info."""
        with patch("iconfucius.tokens.discover_tokens",
                    return_value=self._FAKE_TOKENS):
            with patch("iconfucius.config.get_btc_to_usd_rate",
                        return_value=100000.0):
                result = execute_tool("token_discover", {})
        assert "AlphaToken" in result["display"]
        assert "ALPHA" in result["display"]
        assert "abc1" in result["display"]


class TestTokenPriceExecutor:
    """Tests for the token_price agent skill."""

    _FAKE_API = {
        "id": "29m8",
        "name": "IConfucius",
        "ticker": "ICONFUCIUS",
        "price": 1500,
        "price_5m": 1500,
        "price_1h": 1400,
        "price_6h": 2000,
        "price_1d": 1000,
        "marketcap": 31500000000,
        "volume_24": 20000000000,
        "holder_count": 253,
        "btc_liquidity": 11000000000,
        "divisibility": 8,
        "bonded": True,
    }

    def test_returns_price_data(self):
        """Verify returns price data."""
        with patch("iconfucius.tokens._search_api", return_value=[]):
            with patch("iconfucius.tokens.fetch_token_data",
                        return_value=self._FAKE_API):
                with patch("iconfucius.config.get_btc_to_usd_rate",
                            return_value=100000.0):
                    result = execute_tool("token_price",
                                          {"query": "IConfucius"})
        assert result["status"] == "ok"
        assert result["token_id"] == "29m8"
        assert result["price_sats"] == 1.5  # 1500 msat / 1000
        assert result["ticker"] == "ICONFUCIUS"
        assert "IConfucius" in result["display"]

    def test_price_change_percentages(self):
        """Verify price change percentages."""
        with patch("iconfucius.tokens._search_api", return_value=[]):
            with patch("iconfucius.tokens.fetch_token_data",
                        return_value=self._FAKE_API):
                with patch("iconfucius.config.get_btc_to_usd_rate",
                            return_value=100000.0):
                    result = execute_tool("token_price", {"query": "29m8"})
        # 1500 vs 1400 = +7.1%
        assert result["change_1h"] == "+7.1%"
        # 1500 vs 2000 = -25.0%
        assert result["change_6h"] == "-25.0%"
        # 1500 vs 1000 = +50.0%
        assert result["change_24h"] == "+50.0%"

    def test_missing_query_returns_error(self):
        """Verify missing query returns error."""
        result = execute_tool("token_price", {})
        assert result["status"] == "error"
        assert "required" in result["error"].lower()

    def test_unknown_token_returns_error(self):
        """Verify unknown token returns error."""
        with patch("iconfucius.tokens._search_api", return_value=[]):
            result = execute_tool("token_price",
                                  {"query": "nonexistent_xyz_999"})
        assert result["status"] == "error"
        assert "not found" in result["error"].lower()

    def test_api_failure_returns_error(self):
        """Verify api failure returns error."""
        with patch("iconfucius.tokens._search_api", return_value=[]):
            with patch("iconfucius.tokens.fetch_token_data",
                        return_value=None):
                result = execute_tool("token_price",
                                      {"query": "IConfucius"})
        assert result["status"] == "error"
        assert "Could not fetch" in result["error"]

    def test_usd_rate_failure_graceful(self):
        """Verify usd rate failure graceful."""
        with patch("iconfucius.tokens._search_api", return_value=[]):
            with patch("iconfucius.tokens.fetch_token_data",
                        return_value=self._FAKE_API):
                with patch("iconfucius.config.get_btc_to_usd_rate",
                            side_effect=Exception("offline")):
                    result = execute_tool("token_price",
                                          {"query": "IConfucius"})
        assert result["status"] == "ok"
        assert result["price_sats"] == 1.5  # 1500 msat / 1000
        assert result["price_usd"] is None
        assert "sats" in result["display"]


class TestTokensToSubunits:
    """Tests for human-readable tokens → raw sub-units conversion."""

    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"divisibility": 8})
    def test_basic_conversion(self, _mock):
        # 1000 tokens with div=8 → 100_000_000_000
        """Verify basic conversion."""
        result = _tokens_to_subunits(1000.0, "29m8")
        assert result == 100_000_000_000

    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"divisibility": 0})
    def test_divisibility_zero(self, _mock):
        # div=0 → tokens are indivisible, 1:1
        """Verify divisibility zero."""
        result = _tokens_to_subunits(500.0, "abc")
        assert result == 500

    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"divisibility": 2})
    def test_divisibility_two(self, _mock):
        """Verify divisibility two."""
        result = _tokens_to_subunits(10.5, "xyz")
        assert result == 1050

    @patch("iconfucius.tokens.fetch_token_data", return_value=None)
    def test_missing_data_defaults_to_8(self, _mock):
        """Verify missing data defaults to 8."""
        result = _tokens_to_subunits(1.0, "unknown")
        assert result == 100_000_000

    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"divisibility": 8})
    def test_fractional_tokens(self, _mock):
        # 0.5 tokens with div=8 → 50_000_000
        """Verify fractional tokens."""
        result = _tokens_to_subunits(0.5, "29m8")
        assert result == 50_000_000


class TestUsdConversion:
    """Tests for USD-to-sats and USD-to-tokens conversion."""

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    def test_usd_to_sats(self, _mock):
        # $1 at $100k/BTC = 1000 sats
        """Verify usd to sats."""
        assert _usd_to_sats(1.0) == 1000

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    def test_usd_to_sats_twenty_dollars(self, _mock):
        # $20 at $100k/BTC = 20,000 sats
        """Verify usd to sats twenty dollars."""
        assert _usd_to_sats(20.0) == 20000

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"price": 1500, "divisibility": 8})
    def test_usd_to_tokens(self, _mock_fetch, _mock_usd):
        # $5 at $100k/BTC = 5000 sats
        # raw_tokens = 5000 * 1_000_000 * 10^8 / 1500
        """Verify usd to tokens."""
        tokens = _usd_to_tokens(5.0, "29m8")
        assert tokens > 0
        # Verify: value_sats = (tokens * 1500) / 10^8 / 10^6 ≈ 5000
        value_sats = (tokens * 1500) / (10 ** 8) / 1_000_000
        assert abs(value_sats - 5000) < 1  # within 1 sat

    @patch("iconfucius.tokens.fetch_token_data", return_value=None)
    def test_usd_to_tokens_no_price_raises(self, _mock):
        """Verify usd to tokens no price raises."""
        from pytest import raises
        with raises(ValueError, match="Could not fetch price"):
            _usd_to_tokens(5.0, "nonexistent")


class TestTradeUsdAmount:
    """Tests for amount_usd parameter in trade_buy and trade_sell."""

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    def test_buy_with_amount_usd(self, _mock_usd, tmp_path, monkeypatch):
        """trade_buy with amount_usd converts to sats."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        fake_results = [("bot-1", {"status": "ok", "action": "buy"})]
        with patch("iconfucius.config.require_wallet", return_value=True):
            with patch("iconfucius.cli.concurrent.run_per_bot",
                        return_value=fake_results):
                result = execute_tool("trade_buy", {
                    "token_id": "29m8",
                    "amount_usd": 20.0,
                    "bot_name": "bot-1",
                })
        assert result["status"] == "ok"
        assert result["succeeded"] == 1

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"price": 1500, "divisibility": 8})
    def test_sell_with_amount_usd(self, _mock_fetch, _mock_usd,
                                   tmp_path, monkeypatch):
        """trade_sell with amount_usd converts to raw tokens."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        fake_results = [("bot-1", {"status": "ok", "action": "sell"})]
        with patch("iconfucius.config.require_wallet", return_value=True):
            with patch("iconfucius.cli.concurrent.run_per_bot",
                        return_value=fake_results):
                result = execute_tool("trade_sell", {
                    "token_id": "29m8",
                    "amount_usd": 5.0,
                    "bot_name": "bot-1",
                })
        assert result["status"] == "ok"
        assert result["succeeded"] == 1

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    def test_buy_usd_does_not_mutate_args(self, _mock_usd, tmp_path, monkeypatch):
        """trade_buy with amount_usd must not write back into args dict."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        fake_results = [("bot-1", {"status": "ok", "action": "buy"})]
        args = {"token_id": "29m8", "amount_usd": 20.0, "bot_name": "bot-1"}
        with patch("iconfucius.config.require_wallet", return_value=True):
            with patch("iconfucius.cli.concurrent.run_per_bot",
                        return_value=fake_results):
                execute_tool("trade_buy", args)
        assert "amount" not in args

    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100000.0)
    @patch("iconfucius.tokens.fetch_token_data",
           return_value={"price": 1500, "divisibility": 8})
    def test_sell_usd_does_not_mutate_args(self, _mock_fetch, _mock_usd,
                                            tmp_path, monkeypatch):
        """trade_sell with amount_usd must not write raw subunits into args."""
        monkeypatch.setenv("ICONFUCIUS_ROOT", str(tmp_path))
        fake_results = [("bot-1", {"status": "ok", "action": "sell"})]
        args = {"token_id": "29m8", "amount_usd": 5.0, "bot_name": "bot-1"}
        with patch("iconfucius.config.require_wallet", return_value=True):
            with patch("iconfucius.cli.concurrent.run_per_bot",
                        return_value=fake_results):
                execute_tool("trade_sell", args)
        assert "amount" not in args

    def test_buy_no_amount_returns_error(self):
        """trade_buy without amount or amount_usd returns error."""
        with patch("iconfucius.config.require_wallet", return_value=True):
            result = execute_tool("trade_buy", {
                "token_id": "29m8",
                "bot_name": "bot-1",
            })
        assert result["status"] == "error"

    def test_sell_no_amount_returns_error(self):
        """trade_sell without amount or amount_usd returns error."""
        with patch("iconfucius.config.require_wallet", return_value=True):
            result = execute_tool("trade_sell", {
                "token_id": "29m8",
                "bot_name": "bot-1",
            })
        assert result["status"] == "error"

    @patch("iconfucius.config.get_btc_to_usd_rate",
           side_effect=Exception("offline"))
    def test_buy_usd_conversion_failure(self, _mock):
        """trade_buy returns error when USD conversion fails."""
        with patch("iconfucius.config.require_wallet", return_value=True):
            result = execute_tool("trade_buy", {
                "token_id": "29m8",
                "amount_usd": 20.0,
                "bot_name": "bot-1",
            })
        assert result["status"] == "error"
        assert "USD conversion failed" in result["error"]


class TestAggregateTradeResults:
    """Tests for _aggregate_trade_results helper."""

    def test_all_succeeded(self):
        """Verify all succeeded."""
        from iconfucius.skills.executor import _aggregate_trade_results
        results = [
            ("bot-1", {"status": "ok", "action": "buy", "amount": 3000}),
            ("bot-2", {"status": "ok", "action": "buy", "amount": 2500}),
        ]
        r = _aggregate_trade_results(results, "buy", "29m8")
        assert r["status"] == "ok"
        assert r["succeeded"] == 2
        assert r["failed"] == 0
        assert r["skipped"] == 0
        assert "Bought 29m8 from 2 bot(s)" in r["display"]
        assert r["details"] == [
            {"bot": "bot-1", "amount": 3000},
            {"bot": "bot-2", "amount": 2500},
        ]
        assert r["notes"] == []

    def test_mixed_results(self):
        """Verify mixed results."""
        from iconfucius.skills.executor import _aggregate_trade_results
        results = [
            ("bot-1", {"status": "ok", "action": "sell"}),
            ("bot-2", {"status": "skipped", "reason": "No tokens"}),
            ("bot-3", Exception("connection timeout")),
        ]
        r = _aggregate_trade_results(results, "sell", "29m8")
        assert r["status"] == "partial"
        assert r["succeeded"] == 1
        assert r["failed"] == 1
        assert r["skipped"] == 1

    def test_notes_from_capped_buy(self):
        """Verify notes from capped buy."""
        from iconfucius.skills.executor import _aggregate_trade_results
        results = [
            ("bot-1", {"status": "ok", "action": "buy", "amount": 7371,
                       "note": "Requested 7,380 sats but bot-1 only had 7,371 sats on Odin.Fun. Buy amount was auto-capped."}),
        ]
        r = _aggregate_trade_results(results, "buy", "2r74")
        assert r["succeeded"] == 1
        assert r["details"] == [{"bot": "bot-1", "amount": 7371}]
        assert len(r["notes"]) == 1
        assert "auto-capped" in r["notes"][0]

    def test_all_skipped(self):
        """Verify all skipped."""
        from iconfucius.skills.executor import _aggregate_trade_results
        results = [
            ("bot-1", {"status": "skipped", "reason": "No tokens"}),
        ]
        r = _aggregate_trade_results(results, "sell", "29m8")
        assert r["status"] == "ok"
        assert r["succeeded"] == 0
        assert r["skipped"] == 1


class TestCheckUpdate:
    """Tests for the check_update tool handler."""

    def test_returns_no_update_when_cache_empty(self):
        """Verify returns no update when cache empty."""
        from iconfucius.skills.executor import _update_cache
        _update_cache.clear()
        result = execute_tool("check_update", {})
        assert result["status"] == "ok"
        assert result["update_available"] is False
        assert result["latest_version"] is None
        assert result["upgrade_command"] == "/upgrade"

    def test_returns_update_when_cache_populated(self):
        """Verify returns update when cache populated."""
        from iconfucius.skills.executor import _update_cache
        _update_cache.clear()
        _update_cache["latest_version"] = "99.0.0"
        _update_cache["release_notes"] = "- New feature"
        try:
            result = execute_tool("check_update", {})
            assert result["status"] == "ok"
            assert result["update_available"] is True
            assert result["latest_version"] == "99.0.0"
            assert result["release_notes"] == "- New feature"
            assert result["upgrade_command"] == "/upgrade"
        finally:
            _update_cache.clear()

    def test_includes_running_version(self):
        """Verify includes running version."""
        from iconfucius import __version__
        from iconfucius.skills.executor import _update_cache
        _update_cache.clear()
        result = execute_tool("check_update", {})
        assert result["running_version"] == __version__


class TestAccountLookupExecutor:
    def test_missing_address_returns_error(self):
        """Verify missing address returns error."""
        result = execute_tool("account_lookup", {})
        assert result["status"] == "error"
        assert "Address is required" in result["error"]

    @patch("iconfucius.accounts.lookup_odin_account", return_value=None)
    def test_unknown_address_returns_not_found(self, _mock_lookup):
        """Verify unknown address returns not found."""
        result = execute_tool("account_lookup", {"address": "zzzzz-zzzzz"})
        assert result["status"] == "ok"
        assert result["found"] is False
        assert "No Odin.fun account found" in result["display"]

    @patch("iconfucius.accounts.lookup_odin_account")
    def test_found_account_returns_details(self, mock_lookup):
        """Verify found account returns details."""
        mock_lookup.return_value = {
            "principal": "abc-def-ghi",
            "username": "trader42",
            "btc_wallet_address": "bc1qfake",
            "btc_deposit_address": "bc1qdeposit",
            "bio": "I trade stuff",
            "avatar": None,
            "admin": False,
            "verified": True,
            "follower_count": 100,
            "following_count": 50,
            "created_at": "2024-01-01",
        }
        result = execute_tool("account_lookup", {"address": "abc-def-ghi"})
        assert result["status"] == "ok"
        assert result["found"] is True
        assert result["principal"] == "abc-def-ghi"
        assert result["username"] == "trader42"
        assert "trader42" in result["display"]
        assert "bc1qfake" in result["display"]
        assert "bc1qdeposit" in result["display"]
        assert "100" in result["display"]

    @patch("iconfucius.accounts.lookup_odin_account")
    def test_resolves_btc_address(self, mock_lookup):
        """Verify resolves btc address."""
        mock_lookup.return_value = {
            "principal": "resolved-principal",
            "username": None,
            "btc_wallet_address": "bc1qfake",
            "btc_deposit_address": None,
            "bio": None,
            "avatar": None,
            "admin": False,
            "verified": False,
            "follower_count": 0,
            "following_count": 0,
            "created_at": None,
        }
        result = execute_tool("account_lookup", {"address": "bc1qfake"})
        assert result["status"] == "ok"
        assert result["found"] is True
        assert result["principal"] == "resolved-principal"


# ---------------------------------------------------------------------------
# Tool definition content — ensures AI sees critical constraints
# ---------------------------------------------------------------------------


class TestWalletSendDefinition:
    """Verify wallet_send tool definition contains the BTC minimum and dual-mode info."""

    def _get_wallet_send_def(self):
        """Return the wallet_send tool definition."""
        from iconfucius.skills.definitions import TOOLS
        for t in TOOLS:
            if t["name"] == "wallet_send":
                return t
        raise AssertionError("wallet_send not found in TOOL_DEFINITIONS")

    def test_description_mentions_btc_minimum(self):
        """AI must see the 50,000 sats minimum for BTC sends."""
        desc = self._get_wallet_send_def()["description"]
        assert "50,000" in desc

    def test_description_mentions_ic_principal(self):
        """AI must know it can send ckBTC to an IC principal."""
        desc = self._get_wallet_send_def()["description"]
        assert "principal" in desc.lower()

    def test_description_mentions_btc_address(self):
        """AI must know BTC address mode converts via ckBTC minter."""
        desc = self._get_wallet_send_def()["description"]
        assert "bc1" in desc
        assert "minter" in desc.lower()

    def test_description_forbids_below_minimum(self):
        """AI must see NEVER call with less than 50,000 sats."""
        desc = self._get_wallet_send_def()["description"]
        assert "NEVER" in desc
        assert "50,000" in desc

    def test_only_address_required(self):
        """amount is optional (amount_usd is an alternative), only address required."""
        schema = self._get_wallet_send_def()["input_schema"]
        assert schema["required"] == ["address"]

    def test_amount_field_mentions_btc_minimum(self):
        """amount field reminds AI of the 50,000 minimum for BTC addresses."""
        props = self._get_wallet_send_def()["input_schema"]["properties"]
        assert "50000" in props["amount"]["description"]

    def test_amount_usd_field_mentions_btc_minimum(self):
        """amount_usd field reminds AI of the 50,000 minimum for BTC addresses."""
        props = self._get_wallet_send_def()["input_schema"]["properties"]
        assert "50000" in props["amount_usd"]["description"]

    def test_address_field_describes_both_modes(self):
        """Address field describes both IC principal and BTC address."""
        props = self._get_wallet_send_def()["input_schema"]["properties"]
        addr_desc = props["address"]["description"]
        assert "principal" in addr_desc.lower()
        assert "bc1" in addr_desc

    def test_description_mentions_wallet_monitor(self):
        """AI should know to use wallet_monitor after a BTC send."""
        desc = self._get_wallet_send_def()["description"]
        assert "wallet_monitor" in desc


class TestWalletSendMonitorHintStripped:
    """Verify wallet_send strips the CLI-specific monitor command."""

    @patch("iconfucius.config.require_wallet", return_value=_FAKE_PEM)
    def test_cli_monitor_hint_removed(self, mock_wallet):
        # Simulate CLI output that includes the monitor hint
        """Verify cli monitor hint removed."""
        mock_result = MagicMock()
        mock_result.exit_code = 0
        mock_result.output = (
            "BTC withdrawal initiated! Block index: 123\n"
            "BTC will arrive after the transaction is confirmed.\n"
            "Check progress with: iconfucius wallet balance --monitor\n"
            "Wallet balance: 5,000 sats"
        )

        with patch("typer.testing.CliRunner") as MockRunner:
            MockRunner.return_value.invoke.return_value = mock_result
            result = execute_tool("wallet_send", {
                "amount": "50000", "address": "bc1qfake",
            })

        assert result["status"] == "ok"
        assert "iconfucius wallet balance --monitor" not in result["display"]
        assert "BTC withdrawal initiated" in result["display"]
        assert result["hint"] == "Use wallet_monitor to check withdrawal progress."

    @patch("iconfucius.config.require_wallet", return_value=_FAKE_PEM)
    def test_ckbtc_send_no_hint(self, mock_wallet):
        """ckBTC sends (no BTC withdrawal) should not get a hint."""
        mock_result = MagicMock()
        mock_result.exit_code = 0
        mock_result.output = "Transfer succeeded! Block index: 456"

        with patch("typer.testing.CliRunner") as MockRunner:
            MockRunner.return_value.invoke.return_value = mock_result
            result = execute_tool("wallet_send", {
                "amount": "5000", "address": "rrkah-fqaaa-aaaaa-aaaaq-cai",
            })

        assert result["status"] == "ok"
        assert "hint" not in result


class TestHowToFundWalletHandler:
    """Tests for the how_to_fund_wallet executor handler."""

    def test_no_wallet_returns_error(self):
        """Verify no wallet returns error."""
        with patch("iconfucius.config.require_wallet", return_value=False):
            result = execute_tool("how_to_fund_wallet", {})
        assert result["status"] == "error"
        assert "wallet" in result["error"].lower()

    @patch("iconfucius.transfers.get_btc_address", return_value="bc1qfund123")
    @patch("iconfucius.transfers.create_ckbtc_minter")
    @patch("iconfucius.transfers.get_balance", return_value=0)
    @patch("iconfucius.transfers.create_icrc1_canister")
    @patch("icp_agent.Agent")
    @patch("icp_agent.Client")
    @patch("icp_identity.Identity")
    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100_000.0)
    def test_returns_structured_data(self, mock_rate, MockId, MockClient,
                                      MockAgent, mock_icrc1, mock_bal,
                                      mock_minter, mock_btc_addr, odin_project):
        """Verify returns structured data."""
        mock_identity = MagicMock()
        mock_identity.sender.return_value = MagicMock(
            __str__=lambda s: "fund-principal"
        )
        MockId.from_pem.return_value = mock_identity
        MockId.return_value = MagicMock()

        result = execute_tool("how_to_fund_wallet", {})
        assert result["status"] == "ok"
        assert result["wallet_principal"] == "fund-principal"
        assert result["btc_deposit_address"] == "bc1qfund123"
        assert result["ckbtc_balance_sats"] == 0

    @patch("iconfucius.transfers.get_btc_address", return_value="bc1qfund123")
    @patch("iconfucius.transfers.create_ckbtc_minter")
    @patch("iconfucius.transfers.get_balance", return_value=0)
    @patch("iconfucius.transfers.create_icrc1_canister")
    @patch("icp_agent.Agent")
    @patch("icp_agent.Client")
    @patch("icp_identity.Identity")
    @patch("iconfucius.config.get_btc_to_usd_rate", return_value=100_000.0)
    def test_display_has_both_options(self, mock_rate, MockId, MockClient,
                                      MockAgent, mock_icrc1, mock_bal,
                                      mock_minter, mock_btc_addr, odin_project):
        """Verify display has both options."""
        mock_identity = MagicMock()
        mock_identity.sender.return_value = MagicMock(
            __str__=lambda s: "fund-principal"
        )
        MockId.from_pem.return_value = mock_identity
        MockId.return_value = MagicMock()

        result = execute_tool("how_to_fund_wallet", {})
        display = result["display"]
        assert "Option 1" in display
        assert "Option 2" in display
        assert "bc1qfund123" in display
        assert "fund-principal" in display
        assert "~6 Bitcoin confirmations" in display



"""
iconfucius.cli.transfer — Transfer tokens between Odin.Fun accounts

Transfers tokens from a bot's Odin.Fun account to another Odin.Fun account.
This is an internal platform transfer — no selling/buying involved.

WARNING: Transfers are irreversible. Sending to a wrong address means
permanent loss of tokens.

Fee: 100 sats BTC (deducted from sender's Odin.Fun BTC balance).

Usage:
  iconfucius --bot bot-1 transfer 29m8 all dxqin-ibe62-...
  iconfucius --all-bots transfer 29m8 1000000000 dxqin-ibe62-...
"""

import json
import os

from icp_agent import Agent, Client
from icp_canister import Canister

from iconfucius.accounts import resolve_odin_account
from iconfucius.candid import ODIN_TRADING_CANDID
from iconfucius.config import (
    IC_HOST,
    ODIN_TRADING_CANISTER_ID,
    fmt_tokens,
    get_verify_certificates,
    log,
    require_wallet,
    set_verbose,
)
from iconfucius.siwb import load_session, siwb_login
from iconfucius.tokens import fetch_token_data
from iconfucius.transfers import patch_delegate_sender, unwrap_canister_result


# ---------------------------------------------------------------------------
# Main
# ---------------------------------------------------------------------------

def run_transfer(bot_name: str, token_id: str, amount: str,
                 to_address: str, verbose: bool = False) -> dict:
    """Transfer tokens from a bot's Odin.Fun account to another account.

    Returns a structured dict:
        {"status": "ok"/"error", ...}

    Args:
        bot_name: Name of the source bot.
        token_id: Token ID to transfer (e.g. '29m8').
        amount: Raw token sub-units, or 'all' for entire balance.
        to_address: Destination address — IC principal, BTC deposit address
            (btc_deposit_address), or BTC wallet address (btc_wallet_address)
            of a bot's Odin.fun account. All three are resolved to the
            account's IC principal via the Odin.fun /v1/users search API.
        verbose: If True, enable detailed logging.
    """
    from iconfucius.logging_config import get_logger
    logger = get_logger()

    set_verbose(verbose)
    if not require_wallet():
        return {"status": "error", "error": "No wallet found. Run: iconfucius wallet create"}

    # Fetch token info for display
    token_info = fetch_token_data(token_id)
    ticker = token_info.get("ticker", token_id) if token_info else token_id
    token_label = f"{token_id} ({ticker})" if ticker != token_id else token_id

    # -----------------------------------------------------------------------
    # Step 1: Resolve and validate destination address
    # -----------------------------------------------------------------------
    logger.info("Step 1: Resolve destination address...")

    # If to_address is a configured bot name, look up its principal from
    # the session cache so the AI can pass bot names directly.
    from iconfucius.config import get_bot_names
    if to_address in get_bot_names():
        from iconfucius.siwb import _session_path
        session_file = _session_path(to_address)
        if os.path.exists(session_file):
            with open(session_file) as _f:
                _session = json.load(_f)
            principal = _session.get("bot_principal_text")
            if principal:
                log(f"  Resolved bot name {to_address} -> {principal}")
                to_address = principal
            else:
                return {
                    "status": "error",
                    "error": (
                        f"Session cache for {to_address} has no principal. "
                        f"Run a balance check on {to_address} first."
                    ),
                }
        else:
            return {
                "status": "error",
                "error": (
                    f"No session cache for {to_address}. "
                    f"Run a balance check on {to_address} first to create one."
                ),
            }

    resolved_principal = resolve_odin_account(to_address)
    if not resolved_principal:
        return {
            "status": "error",
            "error": (
                f"'{to_address}' is not a registered Odin.fun account. "
                "Sending to an unknown address would result in permanent loss."
            ),
        }
    if resolved_principal != to_address:
        log(f"  Resolved {to_address} -> {resolved_principal}")
    logger.info("Step 1: Destination validated as Odin.fun account (%s)", resolved_principal)

    # -----------------------------------------------------------------------
    # Step 2: SIWB login (source bot)
    # -----------------------------------------------------------------------
    logger.info("Step 2: SIWB Login (bot=%s)...", bot_name)
    auth = load_session(bot_name=bot_name, verbose=verbose)
    if not auth:
        log("")
        log("No valid cached session, performing full SIWB login...")
        auth = siwb_login(bot_name=bot_name, verbose=verbose)
        set_verbose(verbose)

    delegate_identity = auth["delegate_identity"]
    bot_principal_text = auth["bot_principal_text"]
    patch_delegate_sender(delegate_identity)
    logger.info("Step 2: SIWB Login done")
    log(f"  Bot principal: {bot_principal_text}")

    # Reject self-transfer
    if bot_principal_text == resolved_principal:
        return {
            "status": "error",
            "error": "Cannot transfer to the same account (source = destination).",
        }

    client = Client(url=IC_HOST)
    auth_agent = Agent(delegate_identity, client)

    odin = Canister(
        agent=auth_agent,
        canister_id=ODIN_TRADING_CANISTER_ID,
        candid_str=ODIN_TRADING_CANDID,
    )

    # -----------------------------------------------------------------------
    # Step 3: Check BTC balance for transfer fee
    # -----------------------------------------------------------------------
    TRANSFER_FEE_SATS = 100
    MSAT_PER_SAT = 1000
    logger.info("Step 3: Check BTC balance for transfer fee (bot=%s)...", bot_name)

    btc_balance_msat = unwrap_canister_result(
        odin.getBalance(bot_principal_text, "btc",
                        verify_certificate=get_verify_certificates())
    )
    btc_balance_sats = btc_balance_msat // MSAT_PER_SAT
    logger.info("Step 3: BTC balance: %d sats", btc_balance_sats)

    if btc_balance_sats < TRANSFER_FEE_SATS:
        return {
            "status": "error",
            "error_type": "insufficient_btc_for_fee",
            "error": (
                f"{bot_name} has {btc_balance_sats} sats BTC but transfer "
                f"requires {TRANSFER_FEE_SATS} sats fee."
            ),
            "bot_name": bot_name,
            "btc_balance_sats": btc_balance_sats,
            "fee_sats": TRANSFER_FEE_SATS,
            "token_id": token_id,
            "token_label": token_label,
            "options": [
                "Fund bot from wallet (minimum 5,000 sats deposit) using the fund tool",
                "Sell some tokens first to free up BTC for the fee",
            ],
        }

    # -----------------------------------------------------------------------
    # Step 4: Check token balance
    # -----------------------------------------------------------------------
    logger.info("Step 4: Check token balance (bot=%s)...", bot_name)

    token_before = unwrap_canister_result(
        odin.getBalance(bot_principal_text, token_id,
                             verify_certificate=get_verify_certificates())
    )
    logger.info("Step 4: %s balance: %s", token_label, fmt_tokens(token_before, token_id))

    # -----------------------------------------------------------------------
    # Step 5: Resolve amount
    # -----------------------------------------------------------------------
    transfer_all = amount.lower() == "all"
    if transfer_all:
        amount_int = token_before
        if amount_int <= 0:
            return {
                "status": "error",
                "error": f"No {token_label} to transfer (balance is 0).",
            }
    else:
        amount_int = int(amount)

    if amount_int <= 0:
        return {"status": "error", "error": "Transfer amount must be greater than 0."}

    if amount_int > token_before:
        return {
            "status": "error",
            "error": (
                f"Insufficient {token_label} balance. "
                f"Have: {fmt_tokens(token_before, token_id)}, "
                f"requested: {fmt_tokens(amount_int, token_id)}"
            ),
        }

    # -----------------------------------------------------------------------
    # Step 6: Execute transfer
    # -----------------------------------------------------------------------
    logger.info("Step 6: token_transfer %s %s -> %s...",
                fmt_tokens(amount_int, token_id), token_label, resolved_principal)

    transfer_request = {
        "amount": amount_int,
        "to": resolved_principal,
        "tokenid": token_id,
    }
    log(f"  Transfer request: {transfer_request}")

    result = unwrap_canister_result(
        odin.token_transfer(transfer_request,
                                 verify_certificate=get_verify_certificates())
    )
    log(f"  Result: {result}")

    if isinstance(result, dict) and "err" in result:
        return {"status": "error", "error": str(result["err"])}

    logger.info("Transfer executed successfully (bot=%s)", bot_name)
    return {
        "status": "ok",
        "bot_name": bot_name,
        "token_id": token_id,
        "token_label": token_label,
        "amount": amount_int,
        "to_address": to_address,
        "to_principal": resolved_principal,
        "token_before": token_before,
    }



"""Tests for iconfucius.openai_compat — OpenAI format translator."""

import json

import pytest

from iconfucius.openai_compat import (
    OpenAICompatResponse,
    TextBlock,
    ToolUseBlock,
    _extract_tool_call_from_text,
    anthropic_messages_to_openai,
    anthropic_tools_to_openai,
    openai_response_to_anthropic,
)


class TestAnthropicMessagesToOpenAI:
    """Test anthropic_messages_to_openai()."""

    def test_system_prompt_to_openai(self):
        """Verify system prompt to openai."""
        result = anthropic_messages_to_openai([], "You are helpful.")
        assert result == [{"role": "system", "content": "You are helpful."}]

    def test_empty_system_prompt(self):
        """Verify empty system prompt."""
        result = anthropic_messages_to_openai(
            [{"role": "user", "content": "hi"}], ""
        )
        assert result[0]["role"] == "user"

    def test_text_messages_passthrough(self):
        """Verify text messages passthrough."""
        messages = [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": "Hi there"},
        ]
        result = anthropic_messages_to_openai(messages, "sys")
        assert result[0] == {"role": "system", "content": "sys"}
        assert result[1] == {"role": "user", "content": "Hello"}
        assert result[2] == {"role": "assistant", "content": "Hi there"}

    def test_tool_use_to_openai(self):
        """Verify tool use to openai."""
        messages = [
            {
                "role": "assistant",
                "content": [
                    {"type": "text", "text": "Let me check."},
                    {
                        "type": "tool_use",
                        "id": "call_123",
                        "name": "get_price",
                        "input": {"token": "BTC"},
                    },
                ],
            },
        ]
        result = anthropic_messages_to_openai(messages, "")
        msg = result[0]
        assert msg["role"] == "assistant"
        assert msg["content"] == "Let me check."
        assert len(msg["tool_calls"]) == 1
        tc = msg["tool_calls"][0]
        assert tc["id"] == "call_123"
        assert tc["type"] == "function"
        assert tc["function"]["name"] == "get_price"
        assert json.loads(tc["function"]["arguments"]) == {"token": "BTC"}

    def test_tool_results_to_openai(self):
        """Bundled Anthropic tool results become separate OpenAI tool messages."""
        messages = [
            {
                "role": "user",
                "content": [
                    {
                        "type": "tool_result",
                        "tool_use_id": "call_1",
                        "content": '{"price": 100}',
                    },
                    {
                        "type": "tool_result",
                        "tool_use_id": "call_2",
                        "content": '{"price": 200}',
                    },
                ],
            },
        ]
        result = anthropic_messages_to_openai(messages, "")
        assert len(result) == 2
        assert result[0] == {
            "role": "tool",
            "tool_call_id": "call_1",
            "content": '{"price": 100}',
        }
        assert result[1] == {
            "role": "tool",
            "tool_call_id": "call_2",
            "content": '{"price": 200}',
        }

    def test_cache_control_stripped(self):
        """cache_control blocks in content are stripped (not passed through)."""
        messages = [
            {
                "role": "user",
                "content": [
                    {
                        "type": "text",
                        "text": "Hello",
                        "cache_control": {"type": "ephemeral"},
                    },
                ],
            },
        ]
        result = anthropic_messages_to_openai(messages, "")
        # Should produce a clean user message with text only
        assert result[0]["role"] == "user"
        assert result[0]["content"] == "Hello"

    def test_multiple_tool_use_blocks(self):
        """Multiple tool_use blocks in one assistant message."""
        messages = [
            {
                "role": "assistant",
                "content": [
                    {
                        "type": "tool_use",
                        "id": "c1",
                        "name": "tool_a",
                        "input": {"x": 1},
                    },
                    {
                        "type": "tool_use",
                        "id": "c2",
                        "name": "tool_b",
                        "input": {"y": 2},
                    },
                ],
            },
        ]
        result = anthropic_messages_to_openai(messages, "")
        assert len(result[0]["tool_calls"]) == 2


class TestAnthropicToolsToOpenAI:
    """Test anthropic_tools_to_openai()."""

    def test_tool_definitions_conversion(self):
        """Verify tool definitions conversion."""
        tools = [
            {
                "name": "get_price",
                "description": "Get token price",
                "input_schema": {
                    "type": "object",
                    "properties": {"token": {"type": "string"}},
                    "required": ["token"],
                },
            },
        ]
        result = anthropic_tools_to_openai(tools)
        assert len(result) == 1
        assert result[0]["type"] == "function"
        func = result[0]["function"]
        assert func["name"] == "get_price"
        assert func["description"] == "Get token price"
        assert func["parameters"]["type"] == "object"
        assert "token" in func["parameters"]["properties"]

    def test_cache_control_in_schema_stripped(self):
        """Verify cache control in schema stripped."""
        tools = [
            {
                "name": "t",
                "description": "d",
                "input_schema": {
                    "type": "object",
                    "cache_control": {"type": "ephemeral"},
                },
            },
        ]
        result = anthropic_tools_to_openai(tools)
        assert "cache_control" not in result[0]["function"]["parameters"]

    def test_empty_tools(self):
        """Verify empty tools."""
        assert anthropic_tools_to_openai([]) == []


class TestOpenAIResponseToAnthropic:
    """Test openai_response_to_anthropic()."""

    def test_response_text_only(self):
        """Verify response text only."""
        data = {
            "choices": [
                {"message": {"content": "Hello world", "role": "assistant"}}
            ]
        }
        resp = openai_response_to_anthropic(data)
        assert len(resp.content) == 1
        assert resp.content[0].type == "text"
        assert resp.content[0].text == "Hello world"

    def test_response_with_tool_calls(self):
        """Verify response with tool calls."""
        data = {
            "choices": [
                {
                    "message": {
                        "role": "assistant",
                        "content": None,
                        "tool_calls": [
                            {
                                "id": "call_abc",
                                "type": "function",
                                "function": {
                                    "name": "get_price",
                                    "arguments": '{"token": "BTC"}',
                                },
                            },
                        ],
                    }
                }
            ]
        }
        resp = openai_response_to_anthropic(data)
        assert len(resp.content) == 1
        block = resp.content[0]
        assert block.type == "tool_use"
        assert block.id == "call_abc"
        assert block.name == "get_price"
        assert block.input == {"token": "BTC"}

    def test_response_text_plus_tool_calls(self):
        """Verify response text plus tool calls."""
        data = {
            "choices": [
                {
                    "message": {
                        "role": "assistant",
                        "content": "Let me check.",
                        "tool_calls": [
                            {
                                "id": "call_1",
                                "type": "function",
                                "function": {
                                    "name": "lookup",
                                    "arguments": "{}",
                                },
                            },
                        ],
                    }
                }
            ]
        }
        resp = openai_response_to_anthropic(data)
        assert len(resp.content) == 2
        assert resp.content[0].type == "text"
        assert resp.content[1].type == "tool_use"

    def test_response_empty_choices(self):
        """Verify response empty choices."""
        resp = openai_response_to_anthropic({"choices": []})
        assert len(resp.content) == 1
        assert resp.content[0].type == "text"
        assert resp.content[0].text == ""

    def test_malformed_tool_args(self):
        """Non-JSON arguments string is wrapped in _raw key."""
        data = {
            "choices": [
                {
                    "message": {
                        "role": "assistant",
                        "content": None,
                        "tool_calls": [
                            {
                                "id": "call_x",
                                "type": "function",
                                "function": {
                                    "name": "bad",
                                    "arguments": "not json{{{",
                                },
                            },
                        ],
                    }
                }
            ]
        }
        resp = openai_response_to_anthropic(data)
        block = resp.content[0]
        assert block.type == "tool_use"
        assert block.input == {"_raw": "not json{{{"}


class TestOpenAICompatResponse:
    """Test the response wrapper dataclass."""

    def test_model_dump(self):
        """Verify model dump."""
        resp = OpenAICompatResponse(content=[
            TextBlock(text="hello"),
            ToolUseBlock(id="c1", name="fn", input={"a": 1}),
        ])
        dumped = resp.model_dump(mode="json")
        assert dumped == {
            "content": [
                {"type": "text", "text": "hello"},
                {"type": "tool_use", "id": "c1", "name": "fn",
                 "input": {"a": 1}},
            ]
        }

    def test_model_dump_no_mode(self):
        """Verify model dump no mode."""
        resp = OpenAICompatResponse(content=[TextBlock(text="x")])
        dumped = resp.model_dump()
        assert "content" in dumped

    def test_attribute_access(self):
        """Verify TextBlock and ToolUseBlock support attribute access."""
        tb = TextBlock(text="hi")
        assert tb.type == "text"
        assert tb.text == "hi"

        tub = ToolUseBlock(id="c1", name="fn", input={"k": "v"})
        assert tub.type == "tool_use"
        assert tub.id == "c1"
        assert tub.name == "fn"
        assert tub.input == {"k": "v"}


class TestExtractToolCallFromText:
    """Test _extract_tool_call_from_text() inline tool-call parsing."""

    def test_tool_name_json_pattern(self):
        """Parse 'tool_name {"key": "value"}' shorthand."""
        result = _extract_tool_call_from_text('trade_buy {"token_id": "29m8", "amount_usd": 5}')
        assert result is not None
        assert result.name == "trade_buy"
        assert result.input == {"token_id": "29m8", "amount_usd": 5}

    def test_tool_name_json_with_whitespace(self):
        """Verify tool name json with whitespace."""
        result = _extract_tool_call_from_text('  wallet_balance  {"verbose": true}  ')
        assert result is not None
        assert result.name == "wallet_balance"
        assert result.input == {"verbose": True}

    def test_tool_name_invalid_json_returns_none(self):
        """Verify tool name invalid json returns none."""
        result = _extract_tool_call_from_text('trade_buy {not valid json}')
        assert result is None

    def test_json_object_pattern_still_works(self):
        """Existing {"name": ..., "arguments": ...} pattern still works."""
        text = '{"name": "get_price", "arguments": {"token": "BTC"}}'
        result = _extract_tool_call_from_text(text)
        assert result is not None
        assert result.name == "get_price"
        assert result.input == {"token": "BTC"}

    def test_xml_wrapped_json(self):
        """Verify xml wrapped json."""
        text = '<json>{"name": "trade_buy", "arguments": {"token_id": "29m8"}}</json>'
        result = _extract_tool_call_from_text(text)
        assert result is not None
        assert result.name == "trade_buy"

    def test_no_tool_call_returns_none(self):
        """Verify no tool call returns none."""
        assert _extract_tool_call_from_text("just some plain text") is None

    def test_empty_string_returns_none(self):
        """Verify empty string returns none."""
        assert _extract_tool_call_from_text("") is None
